import datetime
import os.path
import platform
import re
from urllib import parse

import rarfile
from tornado.escape import xhtml_unescape

import sickchill.start
from sickchill import logger, settings
from sickchill.helper.common import try_int

from . import db, helpers, naming

# Address poor support for scgi over unix domain sockets
# this is not nicely handled by python currently
# http://bugs.python.org/issue23636
parse.uses_netloc.append("scgi")

naming_ep_type = (
    "%(seasonnumber)dx%(episodenumber)02d",
    "s%(seasonnumber)02de%(episodenumber)02d",
    "S%(seasonnumber)02dE%(episodenumber)02d",
    "%(seasonnumber)02dx%(episodenumber)02d",
)

sports_ep_type = (
    "%(seasonnumber)dx%(episodenumber)02d",
    "s%(seasonnumber)02de%(episodenumber)02d",
    "S%(seasonnumber)02dE%(episodenumber)02d",
    "%(seasonnumber)02dx%(episodenumber)02d",
)

naming_ep_type_text = ("1x02", "s01e02", "S01E02", "01x02")

naming_multi_ep_type = {
    0: ["-%(episodenumber)02d"] * len(naming_ep_type),
    1: [" - " + x for x in naming_ep_type],
    2: [x + "%(episodenumber)02d" for x in ("x", "e", "E", "x")],
}
naming_multi_ep_type_text = ("extend", "duplicate", "repeat")

naming_sep_type = (" - ", " ")
naming_sep_type_text = (" - ", "space")


def change_https_cert(https_cert):
    """
    Replace HTTPS Certificate file path

    :param https_cert: path to the new certificate file
    :return: True on success, False on failure
    """
    if https_cert == "":
        settings.HTTPS_CERT = ""
        return True

    if os.path.normpath(settings.HTTPS_CERT) != os.path.normpath(https_cert):
        if helpers.makeDir(os.path.dirname(os.path.abspath(https_cert))):
            settings.HTTPS_CERT = os.path.normpath(https_cert)
            logger.info("Changed https cert path to " + https_cert)
        else:
            return False

    return True


def change_https_key(https_key):
    """
    Replace HTTPS Key file path

    :param https_key: path to the new key file
    :return: True on success, False on failure
    """
    if not https_key:
        settings.HTTPS_KEY = ""
        return True

    if os.path.normpath(settings.HTTPS_KEY) != os.path.normpath(https_key):
        if helpers.makeDir(os.path.dirname(os.path.abspath(https_key))):
            settings.HTTPS_KEY = os.path.normpath(https_key)
            logger.info("Changed https key path to " + https_key)
        else:
            return False

    return True


def change_unrar_tool(unrar_tool, unar_tool, bsdtar_tool):
    try:
        rarfile.tool_setup()
    except rarfile.RarCannotExec:
        pass
    else:
        settings.UNAR_TOOL = rarfile.UNRAR_TOOL
        settings.UNAR_TOOL = rarfile.UNAR_TOOL
        settings.BSDTAR_TOOL = rarfile.BSDTAR_TOOL
        return True

    _unrar_tool = rarfile.UNRAR_TOOL
    _unar_tool = rarfile.UNAR_TOOL
    _bsdtar_tool = rarfile.BSDTAR_TOOL

    if unrar_tool and unrar_tool != rarfile.UNRAR_TOOL:
        rarfile.UNRAR_TOOL = unrar_tool

    if unar_tool and unar_tool != rarfile.UNAR_TOOL:
        rarfile.UNAR_TOOL = unar_tool

    if bsdtar_tool and bsdtar_tool != rarfile.BSDTAR_TOOL:
        rarfile.BSDTAR_TOOL = bsdtar_tool

    try:
        rarfile.tool_setup()
    except rarfile.RarCannotExec:
        pass
    else:
        settings.UNAR_TOOL = rarfile.UNRAR_TOOL
        settings.UNAR_TOOL = rarfile.UNAR_TOOL
        settings.BSDTAR_TOOL = rarfile.BSDTAR_TOOL
        return True

    rarfile.UNRAR_TOOL = _unar_tool
    rarfile.UNAR_TOOL = _unar_tool
    rarfile.BSDTAR_TOOL = _bsdtar_tool

    if platform.system() == "Windows":
        # Look for WinRAR installations
        winrar_path = "WinRAR\\UnRAR.exe"
        # Make a set of unique paths to check from existing environment variables
        check_locations = {
            os.path.join(location, winrar_path)
            for location in (
                os.environ.get("ProgramW6432"),
                os.environ.get("ProgramFiles(x86)"),
                os.environ.get("ProgramFiles"),
                re.sub(r"\s?\(x86\)", "", os.environ["ProgramFiles"]),
            )
            if location
        }
        check_locations.add(os.path.join(settings.DATA_DIR, "unrar\\unrar.exe"))

        for check in check_locations:
            if os.path.isfile(check):
                try:
                    rarfile.UNRAR_TOOL = check
                    rarfile.tool_setup()
                    settings.UNRAR_TOOL = check
                    return True
                except rarfile.RarCannotExec:
                    pass

        logger.info("Trying to download unrar.exe and set the path")
        unrar_store = os.path.join(settings.DATA_DIR, "unrar")  # ./unrar (folder)
        unrar_zip = os.path.join(settings.DATA_DIR, "unrar_win.zip")  # file download

        if helpers.download_file("https://sickchill.github.io/unrar/unrar_win.zip", filename=unrar_zip, session=helpers.make_session()) and helpers.extractZip(
            archive=unrar_zip, targetDir=unrar_store
        ):
            try:
                os.remove(unrar_zip)
            except OSError as e:
                logger.info("Unable to delete downloaded file {0}: {1}. You may delete it manually".format(unrar_zip, e.strerror))

            check = os.path.join(unrar_store, "unrar.exe")
            try:
                rarfile.UNRAR_TOOL = check
                rarfile.tool_setup()
                settings.UNRAR_TOOL = check
                logger.info("Successfully downloaded unrar.exe and set as unrar tool")
                return True
            except (rarfile.RarCannotExec, rarfile.RarExecError, OSError, IOError):
                logger.info("Sorry, unrar was not set up correctly. Try installing WinRAR and make sure it is on the system PATH")
        else:
            logger.info("Unable to download unrar.exe")

    # These must always be set to something before returning
    settings.UNRAR_TOOL = rarfile.UNRAR_TOOL = _unrar_tool
    settings.UNAR_TOOL = rarfile.UNAR_TOOL = _unar_tool
    settings.BSDTAR_TOOL = rarfile.BSDTAR_TOOL = _bsdtar_tool

    try:
        rarfile.tool_setup()
        return True
    except rarfile.RarCannotExec:
        if settings.UNPACK == settings.UNPACK_PROCESS_CONTENTS:
            logger.info(_("Disabling UNPACK setting because no unrar/unar/bsdtar is found and accessible in the path or setting."))
            settings.UNPACK = settings.UNPACK_DISABLED
        return False


def change_sickchill_background(background):
    """
    Replace background image file path

    :param background: path to the new background image
    :return: True on success, False on failure
    """
    if not background:
        settings.SICKCHILL_BACKGROUND_PATH = ""
        return True

    background = os.path.normpath(background)
    if not os.path.exists(background):
        logger.info("Background image does not exist: {0}".format(background))
        return False

    settings.SICKCHILL_BACKGROUND_PATH = background

    return True


def change_custom_css(new_css):
    """
    Replace custom css file path

    :param new_css: path to the new css file
    :return: True on success, False on failure
    """
    if not new_css:
        settings.CUSTOM_CSS_PATH = ""
        return True

    new_css = os.path.normpath(new_css)
    if not os.path.isfile(new_css):
        logger.info("Custom css file does not exist: {0}".format(new_css))
        return False
    if not new_css.endswith("css"):
        logger.info("Custom css file should have the .css extension: {0}".format(new_css))
        return False

    settings.CUSTOM_CSS_PATH = new_css
    return True


def change_nzb_dir(nzb_dir):
    """
    Change NZB Folder

    :param nzb_dir: New NZB Folder location
    :return: True on success, False on failure
    """
    if nzb_dir == "":
        settings.NZB_DIR = ""
        return True

    if os.path.normpath(settings.NZB_DIR) != os.path.normpath(nzb_dir):
        if helpers.makeDir(nzb_dir):
            settings.NZB_DIR = os.path.normpath(nzb_dir)
            logger.info("Changed NZB folder to " + nzb_dir)
        else:
            return False

    return True


def change_torrent_dir(torrent_dir):
    """
    Change torrent directory

    :param torrent_dir: New torrent directory
    :return: True on success, False on failure
    """
    if torrent_dir == "":
        settings.TORRENT_DIR = ""
        return True

    if os.path.normpath(settings.TORRENT_DIR) != os.path.normpath(torrent_dir):
        if helpers.makeDir(torrent_dir):
            settings.TORRENT_DIR = os.path.normpath(torrent_dir)
            logger.info("Changed torrent folder to " + torrent_dir)
        else:
            return False

    return True


def change_tv_download_dir(tv_download_dir):
    """
    Change TV_DOWNLOAD directory (used by postprocessor)

    :param tv_download_dir: New tv download directory
    :return: True on success, False on failure
    """
    if tv_download_dir == "":
        settings.TV_DOWNLOAD_DIR = ""
        return True

    if os.path.normpath(settings.TV_DOWNLOAD_DIR) != os.path.normpath(tv_download_dir):
        if helpers.makeDir(tv_download_dir):
            settings.TV_DOWNLOAD_DIR = os.path.normpath(tv_download_dir)
            logger.info("Changed TV download folder to " + tv_download_dir)
        else:
            return False

    return True


def change_unpack_dir(unpack_dir):
    """
    Change UNPACK directory (used by postprocessor)

    :param unpack_dir: New unpack directory
    :return: True on success, False on failure
    """
    if unpack_dir == "":
        settings.UNPACK_DIR = ""
        return True

    if os.path.normpath(settings.UNPACK_DIR) != os.path.normpath(unpack_dir):
        if bool(settings.ROOT_DIRS) and any(helpers.is_subdirectory(unpack_dir, rd) for rd in settings.ROOT_DIRS.split("|")[1:]):
            # don't change if it's in any of the TV root directories
            logger.info("Unable to change unpack directory to a sub-directory of a TV root dir")
            return False

        if helpers.makeDir(unpack_dir):
            settings.UNPACK_DIR = os.path.normpath(unpack_dir)
            logger.info("Changed unpack directory to " + unpack_dir)
        else:
            logger.info("Unable to create unpack directory " + os.path.normpath(unpack_dir) + ", dir not changed.")
            return False

    return True


def change_postprocessor_frequency(freq):
    """
    Change frequency of automatic postprocessing thread

    :param freq: New frequency
    """
    settings.AUTOPOSTPROCESSOR_FREQUENCY = try_int(freq, settings.DEFAULT_AUTOPOSTPROCESSOR_FREQUENCY)

    if settings.AUTOPOSTPROCESSOR_FREQUENCY < settings.MIN_AUTOPOSTPROCESSOR_FREQUENCY:
        settings.AUTOPOSTPROCESSOR_FREQUENCY = settings.MIN_AUTOPOSTPROCESSOR_FREQUENCY

    settings.autoPostProcessorScheduler.cycleTime = datetime.timedelta(minutes=settings.AUTOPOSTPROCESSOR_FREQUENCY)
    return True


def change_daily_search_frequency(freq):
    """
    Change frequency of daily search thread

    :param freq: New frequency
    """
    settings.DAILYSEARCH_FREQUENCY = try_int(freq, settings.DEFAULT_DAILYSEARCH_FREQUENCY)

    if settings.DAILYSEARCH_FREQUENCY < settings.MIN_DAILYSEARCH_FREQUENCY:
        settings.DAILYSEARCH_FREQUENCY = settings.MIN_DAILYSEARCH_FREQUENCY

    settings.dailySearchScheduler.cycleTime = datetime.timedelta(minutes=settings.DAILYSEARCH_FREQUENCY)
    return True


def change_backlog_frequency(freq):
    """
    Change frequency of backlog thread

    :param freq: New frequency
    """
    settings.BACKLOG_FREQUENCY = try_int(freq, settings.DEFAULT_BACKLOG_FREQUENCY)

    settings.MIN_BACKLOG_FREQUENCY = settings.get_backlog_cycle_time()
    if settings.BACKLOG_FREQUENCY < settings.MIN_BACKLOG_FREQUENCY:
        settings.BACKLOG_FREQUENCY = settings.MIN_BACKLOG_FREQUENCY

    settings.backlogSearchScheduler.cycleTime = datetime.timedelta(minutes=settings.BACKLOG_FREQUENCY)
    return True


def change_update_frequency(freq):
    """
    Change frequency of daily updater thread

    :param freq: New frequency
    """
    settings.UPDATE_FREQUENCY = try_int(freq, settings.DEFAULT_UPDATE_FREQUENCY)

    if settings.UPDATE_FREQUENCY < settings.MIN_UPDATE_FREQUENCY:
        settings.UPDATE_FREQUENCY = settings.MIN_UPDATE_FREQUENCY

    settings.versionCheckScheduler.cycleTime = datetime.timedelta(hours=settings.UPDATE_FREQUENCY)
    return True


def change_showupdate_hour(freq):
    """
    Change frequency of show updater thread

    :param freq: New frequency
    """
    settings.SHOWUPDATE_HOUR = try_int(freq, settings.DEFAULT_SHOWUPDATE_HOUR)

    if settings.SHOWUPDATE_HOUR > 23:
        settings.SHOWUPDATE_HOUR = 0
    elif settings.SHOWUPDATE_HOUR < 0:
        settings.SHOWUPDATE_HOUR = 0

    settings.showUpdateScheduler.start_time = datetime.time(hour=settings.SHOWUPDATE_HOUR)
    return True


def change_subtitle_finder_frequency(subtitles_finder_frequency):
    """
    Change frequency of subtitle thread

    :param subtitles_finder_frequency: New frequency
    """
    if subtitles_finder_frequency == "" or subtitles_finder_frequency is None:
        subtitles_finder_frequency = 1

    settings.SUBTITLES_FINDER_FREQUENCY = try_int(subtitles_finder_frequency, 1)
    return True


def change_version_notify(version_notify):
    """
    Enable/Disable versioncheck thread

    :param version_notify: New desired state
    """
    version_notify = checkbox_to_value(version_notify)

    if settings.VERSION_NOTIFY == version_notify:
        return True

    settings.VERSION_NOTIFY = version_notify
    if settings.VERSION_NOTIFY:
        if not settings.versionCheckScheduler.enable:
            logger.info("Starting VERSIONCHECK thread")
            settings.versionCheckScheduler.silent = False
            settings.versionCheckScheduler.enable = True
            settings.versionCheckScheduler.forceRun()
    else:
        settings.versionCheckScheduler.enable = False
        settings.versionCheckScheduler.silent = True
        logger.info("Stopping VERSIONCHECK thread")

    return True


def change_download_propers(download_propers):
    """
    Enable/Disable proper download thread

    :param download_propers: New desired state
    """
    download_propers = checkbox_to_value(download_propers)

    if settings.DOWNLOAD_PROPERS == download_propers:
        return True

    settings.DOWNLOAD_PROPERS = download_propers
    if settings.DOWNLOAD_PROPERS:
        if not settings.properFinderScheduler.enable:
            logger.info("Starting PROPERFINDER thread")
            settings.properFinderScheduler.silent = False
            settings.properFinderScheduler.enable = True
    else:
        settings.properFinderScheduler.enable = False
        settings.properFinderScheduler.silent = True
        logger.info("Stopping PROPERFINDER thread")

    return True


def change_use_trakt(use_trakt):
    """
    Enable/disable trakt thread

    :param use_trakt: New desired state
    """
    use_trakt = checkbox_to_value(use_trakt)

    if settings.USE_TRAKT == use_trakt:
        return True

    settings.USE_TRAKT = use_trakt
    if settings.USE_TRAKT:
        if not settings.traktCheckerScheduler.enable:
            logger.info("Starting TRAKTCHECKER thread")
            settings.traktCheckerScheduler.silent = False
            settings.traktCheckerScheduler.enable = True
    else:
        settings.traktCheckerScheduler.enable = False
        settings.traktCheckerScheduler.silent = True
        logger.info("Stopping TRAKTCHECKER thread")

    return True


def change_use_subtitles(use_subtitles):
    """
    Enable/Disable subtitle searcher

    :param use_subtitles: New desired state
    """
    use_subtitles = checkbox_to_value(use_subtitles)

    if settings.USE_SUBTITLES == use_subtitles:
        return True

    settings.USE_SUBTITLES = use_subtitles
    if settings.USE_SUBTITLES:
        if not settings.subtitlesFinderScheduler.enable:
            logger.info("Starting SUBTITLESFINDER thread")
            settings.subtitlesFinderScheduler.silent = False
            settings.subtitlesFinderScheduler.enable = True
    else:
        settings.subtitlesFinderScheduler.enable = False
        settings.subtitlesFinderScheduler.silent = True
        logger.debug("Stopping SUBTITLESFINDER thread")

    return True


def change_process_automatically(process_automatically):
    """
    Enable/Disable postprocessor thread

    :param process_automatically: New desired state
    """
    process_automatically = checkbox_to_value(process_automatically)

    if settings.PROCESS_AUTOMATICALLY == process_automatically:
        return True

    settings.PROCESS_AUTOMATICALLY = process_automatically
    if settings.PROCESS_AUTOMATICALLY:
        if not settings.autoPostProcessorScheduler.enable:
            logger.info("Starting POSTPROCESSOR thread")
            settings.autoPostProcessorScheduler.silent = False
            settings.autoPostProcessorScheduler.enable = True
    else:
        logger.info("Stopping POSTPROCESSOR thread")
        settings.autoPostProcessorScheduler.enable = False
        settings.autoPostProcessorScheduler.silent = True

    return True


def check_section(cfg, sec):
    """ Check if INI section exists, if not create it """

    if sec in cfg:
        return True

    cfg[sec] = {}
    return False


def checkbox_to_value(option, value_on=True, value_off=False):
    """
    Turns checkbox option 'on' or 'true' to value_on (True)
    any other value returns value_off (False)
    """

    if isinstance(option, list):
        option = option[-1]
    if isinstance(option, str):
        option = option.strip().lower()

    if option in (True, "on", "true", value_on) or try_int(option) > 0:
        return value_on

    return value_off


def clean_host(host, default_port=None):
    """
    Returns host or host:port or empty string from a given url or host
    If no port is found and default_port is given use host:default_port
    """

    host = host.strip()

    if host:

        match_host_port = re.search(r"(?:http.*://)?(?P<host>[^:/]+).?(?P<port>[0-9]*).*", host)

        cleaned_host = match_host_port.group("host")
        cleaned_port = match_host_port.group("port")

        if cleaned_host:

            if cleaned_port:
                host = cleaned_host + ":" + cleaned_port

            elif default_port:
                host = cleaned_host + ":" + str(default_port)

            else:
                host = cleaned_host

        else:
            host = ""

    return host


def clean_hosts(hosts, default_port=None):
    """
    Returns list of cleaned hosts by clean_host

    :param hosts: list of hosts
    :param default_port: default port to use
    :return: list of cleaned hosts
    """
    cleaned_hosts = []

    for cur_host in [host.strip() for host in hosts.split(",") if host.strip()]:
        cleaned_host = clean_host(cur_host, default_port)
        if cleaned_host:
            cleaned_hosts.append(cleaned_host)

    cleaned_hosts = ",".join(cleaned_hosts) if cleaned_hosts else ""

    return cleaned_hosts


def clean_url(url):
    """
    Returns an cleaned url starting with a scheme and folder with trailing /
    or an empty string
    """

    if url and url.strip():
        url = xhtml_unescape(url.strip())

        if "://" not in url:
            url = "//" + url

        scheme, netloc, path, query, fragment = parse.urlsplit(url, "http")
        if not path:
            path += "/"

        cleaned_url = parse.urlunsplit((scheme, netloc, path, query, fragment))

    else:
        cleaned_url = ""

    return cleaned_url


################################################################################
# min_max                                                                      #
################################################################################
def min_max(val, default, low, high):
    """ Return value forced within range """

    val = try_int(val, default)

    if val < low:
        return low
    if val > high:
        return high

    return val


################################################################################
# check_setting_int                                                            #
################################################################################
def check_setting_int(config, cfg_name, item_name, def_val=0, min_val=None, max_val=None, fallback_def=True, silent=True):
    """
    Checks config setting of integer type

    :param config: config object
    :type config: ConfigObj()
    :param cfg_name: section name of config
    :param item_name: item name of section
    :param def_val: default value to return in case a value can't be retrieved from config,
                    or in case value couldn't be converted,
                    or if `value < min_val` or `value > max_val` (default: 0)
    :param min_val: force value to be greater than or equal to `min_val` (optional)
    :param max_val: force value to be lesser than or equal to `max_val` (optional)
    :param fallback_def: if True, `def_val` will be returned when value not in range of `min_val` and `max_val`.
                         otherwise, `min_val`/`max_val` value will be returned respectively (default: True)
    :param silent: don't log result to debug log (default: True)

    :return: value of `config[cfg_name][item_name]` or `min_val`/`max_val` (see def_low_high) `def_val` (see def_val)
    :rtype: int
    """
    if not isinstance(def_val, int):
        logger.error("{dom}:{key} default value is not the correct type. Expected {t}, got {dt}".format(dom=cfg_name, key=item_name, t="int", dt=type(def_val)))

    if not (min_val is None or isinstance(min_val, int)):
        logger.error("{dom}:{key} min_val value is not the correct type. Expected {t}, got {dt}".format(dom=cfg_name, key=item_name, t="int", dt=type(min_val)))

    if not (max_val is None or isinstance(max_val, int)):
        logger.error("{dom}:{key} max_val value is not the correct type. Expected {t}, got {dt}".format(dom=cfg_name, key=item_name, t="int", dt=type(max_val)))

    try:
        if not (check_section(config, cfg_name) and check_section(config[cfg_name], item_name)):
            raise ValueError

        my_val = config[cfg_name][item_name]

        if str(my_val).lower() == "true":
            my_val = 1
        elif str(my_val).lower() == "false":
            my_val = 0

        my_val = int(my_val)

        if isinstance(min_val, int) and my_val < min_val:
            my_val = config[cfg_name][item_name] = (min_val, def_val)[fallback_def]
        if isinstance(max_val, int) and my_val > max_val:
            my_val = config[cfg_name][item_name] = (max_val, def_val)[fallback_def]

    except (ValueError, IndexError, KeyError, TypeError):
        my_val = def_val

        if cfg_name not in config:
            config[cfg_name] = {}

        config[cfg_name][item_name] = my_val

    if not silent:
        logger.debug(item_name + " -> " + str(my_val))

    return my_val


################################################################################
# check_setting_float                                                          #
################################################################################
def check_setting_float(config, cfg_name, item_name, def_val=0.0, min_val=None, max_val=None, fallback_def=True, silent=True):
    """
    Checks config setting of float type

    :param config: config object
    :type config: ConfigObj()
    :param cfg_name: section name of config
    :param item_name: item name of section
    :param def_val: default value to return in case a value can't be retrieved from config
                    or if couldn't be converted,
                    or if `value < min_val` or `value > max_val` (default: 0.0)
    :param min_val: force value to be greater than or equal to `min_val` (optional)
    :param max_val: force value to be lesser than or equal to `max_val` (optional)
    :param fallback_def: if True, `def_val` will be returned when value not in range of `min_val` and `max_val`.
                         otherwise, `min_val`/`max_val` value will be returned respectively (default: True)
    :param silent: don't log result to debug log (default: True)

    :return: value of `config[cfg_name][item_name]` or `min_val`/`max_val` (see def_low_high) `def_val` (see def_val)
    :rtype: float
    """
    if not isinstance(def_val, float):
        logger.error(
            "{dom}:{key} default value is not the correct type. Expected {t}, got {dt}".format(dom=cfg_name, key=item_name, t="float", dt=type(def_val))
        )

    if not (min_val is None or isinstance(min_val, float)):
        logger.error(
            "{dom}:{key} min_val value is not the correct type. Expected {t}, got {dt}".format(dom=cfg_name, key=item_name, t="float", dt=type(min_val))
        )

    if not (max_val is None or isinstance(max_val, float)):
        logger.error(
            "{dom}:{key} max_val value is not the correct type. Expected {t}, got {dt}".format(dom=cfg_name, key=item_name, t="float", dt=type(max_val))
        )

    try:
        if not (check_section(config, cfg_name) and check_section(config[cfg_name], item_name)):
            raise ValueError

        my_val = float(config[cfg_name][item_name])

        if isinstance(min_val, float) and my_val < min_val:
            my_val = config[cfg_name][item_name] = (min_val, def_val)[fallback_def]
        if isinstance(max_val, float) and my_val > max_val:
            my_val = config[cfg_name][item_name] = (max_val, def_val)[fallback_def]
    except (ValueError, IndexError, KeyError, TypeError):
        my_val = def_val

        if cfg_name not in config:
            config[cfg_name] = {}

        config[cfg_name][item_name] = my_val

    if not silent:
        logger.debug(item_name + " -> " + str(my_val))

    return my_val


################################################################################
# check_setting_str                                                            #
################################################################################
def check_setting_str(config, cfg_name, item_name, def_val=str(""), silent=True, censor_log=False):
    """
    Checks config setting of string types

    :param config: config object
    :type config: ConfigObj()
    :param cfg_name: section name of config
    :param item_name: item name of section
    :param def_val: default value to return in case a value can't be retrieved from config
                    or if couldn't be converted (default: empty str)
    :param silent: don't log result to debug log (default: True)
    :param censor_log: overrides and adds this setting to logger censored items (default: False)

    :return: decrypted value of `config[cfg_name][item_name]`
             or `def_val` (see cases of def_val)
    :rtype: str
    """
    if not isinstance(def_val, str):
        logger.error(
            "{dom}:{key} default value is not the correct type. Expected {t}, got {dt}".format(dom=cfg_name, key=item_name, t="string", dt=type(def_val))
        )

    # For passwords you must include the word `password` in the item_name and add `helpers.encrypt(ITEM_NAME, ENCRYPTION_VERSION)` in save_config()
    encryption_version = (0, settings.ENCRYPTION_VERSION)["password" in item_name]

    try:
        if not (check_section(config, cfg_name) and item_name in config[cfg_name]):
            raise ValueError

        my_val = helpers.decrypt(config[cfg_name][item_name], encryption_version)
        if str(my_val) == str(None) or not str(my_val):
            raise ValueError
    except (ValueError, IndexError, KeyError):
        my_val = def_val

        if cfg_name not in config:
            config[cfg_name] = {}

        config[cfg_name][item_name] = helpers.encrypt(my_val, encryption_version)

    if (censor_log or (cfg_name, item_name) in logger.censored_items.items()) and not item_name.endswith("custom_url"):
        logger.censored_items[cfg_name, item_name] = my_val

    if not silent:
        logger.debug(item_name + " -> " + my_val)

    return str(my_val)


################################################################################
# check_setting_bool                                                           #
################################################################################
def check_setting_bool(config, cfg_name, item_name, def_val=False, silent=True):
    """
    Checks config setting of boolean type

    :param config: config object
    :type config: ConfigObj()
    :param cfg_name: section name of config
    :param item_name: item name of section
    :param def_val: default value to return in case a value can't be retrieved from config
                    or if couldn't be converted (default: False)
    :param silent: don't log result to debug log (default: True)

    :return: value of `config[cfg_name][item_name]`
             or `def_val` (see cases of def_val)
    :rtype: bool
    """
    try:
        if not isinstance(def_val, bool):
            logger.error(
                "{dom}:{key} default value is not the correct type. Expected {t}, got {dt}".format(dom=cfg_name, key=item_name, t="bool", dt=type(def_val))
            )

        if not (check_section(config, cfg_name) and item_name in config[cfg_name]):
            raise ValueError

        my_val = config[cfg_name][item_name]
        my_val = str(my_val)
        if my_val == str(None) or not my_val:
            raise ValueError

        my_val = checkbox_to_value(my_val)
    except (KeyError, IndexError, ValueError):
        my_val = bool(def_val)

        if cfg_name not in config:
            config[cfg_name] = {}

        config[cfg_name][item_name] = my_val

    if not silent:
        logger.debug(item_name + " -> " + str(my_val))

    return my_val


class ConfigMigrator(object):
    def __init__(self, config_obj):
        """
        Initializes a config migrator that can take the config from the version indicated in the config
        file up to the version required by SB
        """

        self.config_obj = config_obj

        # check the version of the config
        self.config_version = check_setting_int(config_obj, "General", "config_version", settings.CONFIG_VERSION)
        self.expected_config_version = settings.CONFIG_VERSION
        self.migration_names = {
            1: "Custom naming",
            2: "Sync backup number with version number",
            3: "Rename omgwtfnzb variables",
            4: "Add newznab catIDs",
            5: "Metadata update",
            6: "Convert from XBMC to new KODI variables",
            7: "Use version 2 for password encryption",
            8: "Convert Plex setting keys",
            9: "Rename autopostprocesser (typo) to autopostprocessor",
            10: "Refactor flatten_folders_default to season_folders_default",
            11: "Remove old Kodi notifier, and move Kodi 12+ settings under the correct variables",
            12: "Convert new_qbittorrent and rtorrent9 setting to qbittorent and rtorrent",
        }

    def migrate_config(self):
        """
        Calls each successive migration until the config is the same version as SB expects
        """

        if self.config_version > self.expected_config_version:
            logger.log_error_and_exit(
                """Your config version ({0:d}) has been incremented past what this version of SickChill supports ({1:d}).
                If you have used other forks or a newer version of SickChill, your config file may be unusable due to their modifications.""".format(
                    self.config_version, self.expected_config_version
                )
            )

        settings.CONFIG_VERSION = self.config_version

        while self.config_version < self.expected_config_version:
            next_version = self.config_version + 1

            if next_version in self.migration_names:
                migration_name = ": " + self.migration_names[next_version]
            else:
                migration_name = ""

            logger.info("Backing up config before upgrade")
            if not helpers.backupVersionedFile(settings.CONFIG_FILE, self.config_version):
                logger.log_error_and_exit("Config backup failed, abort upgrading config")
            else:
                logger.info("Proceeding with upgrade")

            # do the migration, expect a method named _migrate_v<num>
            logger.info("Migrating config up to version " + str(next_version) + migration_name)
            getattr(self, "_migrate_v" + str(next_version))()
            self.config_version = next_version

            # save new config after migration
            settings.CONFIG_VERSION = self.config_version
            logger.info("Saving config file to disk")
            sickchill.start.save_config()

    # Migration v1: Custom naming
    def _migrate_v1(self):
        """
        Reads in the old naming settings from your config and generates a new config template from them.
        """

        settings.NAMING_PATTERN = self._name_to_pattern()
        logger.info("Based on your old settings I'm setting your new naming pattern to: " + settings.NAMING_PATTERN)

        settings.NAMING_CUSTOM_ABD = check_setting_bool(self.config_obj, "General", "naming_dates")

        if settings.NAMING_CUSTOM_ABD:
            settings.NAMING_ABD_PATTERN = self._name_to_pattern(True)
            logger.info("Adding a custom air-by-date naming pattern to your config: " + settings.NAMING_ABD_PATTERN)
        else:
            settings.NAMING_ABD_PATTERN = naming.name_abd_presets[0]

        settings.NAMING_MULTI_EP = int(check_setting_int(self.config_obj, "General", "naming_multi_ep_type", 1))

        # see if any of their shows used season folders
        main_db_con = db.DBConnection()
        season_folder_shows = main_db_con.select("SELECT indexer_id FROM tv_shows WHERE flatten_folders = 0 LIMIT 1")

        # if any shows had season folders on then prepend season folder to the pattern
        if season_folder_shows:

            old_season_format = check_setting_str(self.config_obj, "General", "season_folders_format", "Season %02d")

            if old_season_format:
                try:
                    new_season_format = old_season_format % 9
                    new_season_format = str(new_season_format).replace("09", "%0S")
                    new_season_format = new_season_format.replace("9", "%S")

                    logger.info("Changed season folder format from " + old_season_format + " to " + new_season_format + ", prepending it to your naming config")
                    settings.NAMING_PATTERN = new_season_format + os.sep + settings.NAMING_PATTERN

                except (TypeError, ValueError):
                    logger.exception("Can't change " + old_season_format + " to new season format")

        # if no shows had it on then don't flatten any shows and don't put season folders in the config
        else:

            logger.info("No shows were using season folders before so I'm disabling flattening on all shows")

            # don't flatten any shows at all
            main_db_con.action("UPDATE tv_shows SET flatten_folders = 0")

        settings.NAMING_FORCE_FOLDERS = naming.check_force_season_folders()

    def _name_to_pattern(self, abd=False):

        # get the old settings from the file
        use_periods = check_setting_bool(self.config_obj, "General", "naming_use_periods")
        ep_type = check_setting_int(self.config_obj, "General", "naming_ep_type")
        sep_type = check_setting_int(self.config_obj, "General", "naming_sep_type")
        use_quality = check_setting_bool(self.config_obj, "General", "naming_quality")

        use_show_name = check_setting_bool(self.config_obj, "General", "naming_show_name", True)
        use_ep_name = check_setting_bool(self.config_obj, "General", "naming_ep_name", True)

        # make the presets into templates
        _naming_ep_type = ("%Sx%0E", "s%0Se%0E", "S%0SE%0E", "%0Sx%0E")

        # set up our data to use
        if use_periods:
            show_name = "%S.N"
            ep_name = "%E.N"
            ep_quality = "%Q.N"
            abd_string = "%A.D"
        else:
            show_name = "%SN"
            ep_name = "%EN"
            ep_quality = "%QN"
            abd_string = "%A-D"

        if abd and abd_string:
            ep_string = abd_string
        else:
            ep_string = _naming_ep_type[ep_type]

        finalName = ""

        # start with the show name
        if use_show_name and show_name:
            finalName += show_name + naming_sep_type[sep_type]

        # add the season/ep stuff
        finalName += ep_string

        # add the episode name
        if use_ep_name and ep_name:
            finalName += naming_sep_type[sep_type] + ep_name

        # add the quality
        if use_quality and ep_quality:
            finalName += naming_sep_type[sep_type] + ep_quality

        if use_periods:
            finalName = re.sub(r"\s+", ".", finalName)

        return finalName

    # Migration v2: Dummy migration to sync backup number with config version number
    @staticmethod
    def _migrate_v2():
        return

    # Migration v3: Rename omgwtfnzb variables
    def _migrate_v3(self):
        """
        Reads in the old naming settings from your config and generates a new config template from them.
        """
        # get the old settings from the file and store them in the new variable names
        settings.OMGWTFNZBS_USERNAME = check_setting_str(self.config_obj, "omgwtfnzbs", "omgwtfnzbs_uid")
        settings.OMGWTFNZBS_APIKEY = check_setting_str(self.config_obj, "omgwtfnzbs", "omgwtfnzbs_key")

    # Migration v4: Add default newznab catIDs
    def _migrate_v4(self):
        """ Update newznab providers so that the category IDs can be set independently via the config """

        new_newznab_data = []
        old_newznab_data = check_setting_str(self.config_obj, "Newznab", "newznab_data")

        if old_newznab_data:
            old_newznab_data_list = old_newznab_data.split("!!!")

            for cur_provider_data in old_newznab_data_list:
                try:
                    name, url, key, enabled = cur_provider_data.split("|")
                except ValueError:
                    logger.error("Skipping Newznab provider string: '" + cur_provider_data + "', incorrect format")
                    continue

                if name == "Sick Beard Index":
                    key = "0"

                if name == "NZBs.org":
                    catIDs = "5030,5040,5060,5070,5090"
                else:
                    catIDs = "5030,5040,5060"

                cur_provider_data_list = [name, url, key, catIDs, enabled]
                new_newznab_data.append("|".join(cur_provider_data_list))

            settings.NEWZNAB_DATA = "!!!".join(new_newznab_data)

    # Migration v5: Metadata upgrade
    def _migrate_v5(self):
        """ Updates metadata values to the new format """

        """ Quick overview of what the upgrade does:

        new | old | description (new)
        ----+-----+--------------------
          1 |  1  | show metadata
          2 |  2  | episode metadata
          3 |  4  | show fanart
          4 |  3  | show poster
          5 |  -  | show banner
          6 |  5  | episode thumb
          7 |  6  | season poster
          8 |  -  | season banner
          9 |  -  | season all poster
         10 |  -  | season all banner

        Note that the ini places start at 1 while the list index starts at 0.
        old format: 0|0|0|0|0|0 -- 6 places
        new format: 0|0|0|0|0|0|0|0|0|0 -- 10 places

        Drop the use of use_banner option.
        Migrate the poster override to just using the banner option (applies to xbmc only).
        """

        metadata_xbmc = check_setting_str(self.config_obj, "General", "metadata_xbmc", "0|0|0|0|0|0")
        metadata_xbmc_12plus = check_setting_str(self.config_obj, "General", "metadata_xbmc_12plus", "0|0|0|0|0|0")
        metadata_mediabrowser = check_setting_str(self.config_obj, "General", "metadata_mediabrowser", "0|0|0|0|0|0")
        metadata_ps3 = check_setting_str(self.config_obj, "General", "metadata_ps3", "0|0|0|0|0|0")
        metadata_wdtv = check_setting_str(self.config_obj, "General", "metadata_wdtv", "0|0|0|0|0|0")
        metadata_tivo = check_setting_str(self.config_obj, "General", "metadata_tivo", "0|0|0|0|0|0")
        metadata_mede8er = check_setting_str(self.config_obj, "General", "metadata_mede8er", "0|0|0|0|0|0")

        use_banner = check_setting_bool(self.config_obj, "General", "use_banner")

        def _migrate_metadata(metadata, metadata_name, _use_banner):
            cur_metadata = metadata.split("|")
            # if target has the old number of values, do upgrade
            if len(cur_metadata) == 6:
                logger.info("Upgrading " + metadata_name + " metadata, old value: " + metadata)
                cur_metadata.insert(4, "0")
                cur_metadata.append("0")
                cur_metadata.append("0")
                cur_metadata.append("0")
                # swap show fanart, show poster
                cur_metadata[3], cur_metadata[2] = cur_metadata[2], cur_metadata[3]
                # if user was using _use_banner to override the poster, instead enable the banner option and deactivate poster
                if metadata_name == "XBMC" and _use_banner:
                    cur_metadata[4], cur_metadata[3] = cur_metadata[3], "0"
                # write new format
                metadata = "|".join(cur_metadata)
                logger.info("Upgrading " + metadata_name + " metadata, new value: " + metadata)

            elif len(cur_metadata) == 10:

                metadata = "|".join(cur_metadata)
                logger.info("Keeping " + metadata_name + " metadata, value: " + metadata)

            else:
                logger.error("Skipping " + metadata_name + " metadata: '" + metadata + "', incorrect format")
                metadata = "0|0|0|0|0|0|0|0|0|0"
                logger.info("Setting " + metadata_name + " metadata, new value: " + metadata)

            return metadata

        settings.METADATA_XBMC = _migrate_metadata(metadata_xbmc, "XBMC", use_banner)
        settings.METADATA_XBMC_12PLUS = _migrate_metadata(metadata_xbmc_12plus, "XBMC 12+", use_banner)
        settings.METADATA_MEDIABROWSER = _migrate_metadata(metadata_mediabrowser, "MediaBrowser", use_banner)
        settings.METADATA_PS3 = _migrate_metadata(metadata_ps3, "PS3", use_banner)
        settings.METADATA_WDTV = _migrate_metadata(metadata_wdtv, "WDTV", use_banner)
        settings.METADATA_TIVO = _migrate_metadata(metadata_tivo, "TIVO", use_banner)
        settings.METADATA_MEDE8ER = _migrate_metadata(metadata_mede8er, "Mede8er", use_banner)

    # Migration v6: Convert from XBMC to KODI variables
    def _migrate_v6(self):
        settings.USE_KODI = check_setting_bool(self.config_obj, "XBMC", "use_xbmc")
        settings.KODI_ALWAYS_ON = check_setting_bool(self.config_obj, "XBMC", "xbmc_always_on", True)
        settings.KODI_NOTIFY_ONSNATCH = check_setting_bool(self.config_obj, "XBMC", "xbmc_notify_onsnatch")
        settings.KODI_NOTIFY_ONDOWNLOAD = check_setting_bool(self.config_obj, "XBMC", "xbmc_notify_ondownload")
        settings.KODI_NOTIFY_ONSUBTITLEDOWNLOAD = check_setting_bool(self.config_obj, "XBMC", "xbmc_notify_onsubtitledownload")
        settings.KODI_UPDATE_LIBRARY = check_setting_bool(self.config_obj, "XBMC", "xbmc_update_library")
        settings.KODI_UPDATE_FULL = check_setting_bool(self.config_obj, "XBMC", "xbmc_update_full")
        settings.KODI_UPDATE_ONLYFIRST = check_setting_bool(self.config_obj, "XBMC", "xbmc_update_onlyfirst")
        settings.KODI_HOST = check_setting_str(self.config_obj, "XBMC", "xbmc_host")
        settings.KODI_USERNAME = check_setting_str(self.config_obj, "XBMC", "xbmc_username", censor_log=True)
        settings.KODI_PASSWORD = check_setting_str(self.config_obj, "XBMC", "xbmc_password", censor_log=True)
        settings.METADATA_KODI = check_setting_str(self.config_obj, "General", "metadata_xbmc", "0|0|0|0|0|0|0|0|0|0")
        settings.METADATA_KODI_12PLUS = check_setting_str(self.config_obj, "General", "metadata_xbmc_12plus", "0|0|0|0|0|0|0|0|0|0")

    # Migration v7: Use version 2 for password encryption
    @staticmethod
    def _migrate_v7():
        settings.ENCRYPTION_VERSION = 2

    # Migration v8: Rename plex settings
    def _migrate_v8(self):
        settings.PLEX_CLIENT_HOST = check_setting_str(self.config_obj, "Plex", "plex_host")
        settings.PLEX_SERVER_USERNAME = check_setting_str(self.config_obj, "Plex", "plex_username", censor_log=True)
        settings.PLEX_SERVER_PASSWORD = check_setting_str(self.config_obj, "Plex", "plex_password", censor_log=True)
        settings.USE_PLEX_SERVER = check_setting_bool(self.config_obj, "Plex", "use_plex")

    # Migration v9: Rename autopostprocesser (typo) to autopostprocessor
    def _migrate_v9(self):
        settings.AUTOPOSTPROCESSOR_FREQUENCY = check_setting_str(self.config_obj, "General", "autopostprocesser_frequency")

    # Migration v10: Change flatten_folders_default to season_folders_default (inverted)
    def _migrate_v10(self):
        settings.SEASON_FOLDERS_DEFAULT = not check_setting_str(self.config_obj, "General", "flatten_folders_default")

    # Migration v11: Remove old kodi and only use Kodi 12+
    def _migrate_v11(self):
        settings.METADATA_KODI = check_setting_str(settings.CFG, "General", "metadata_kodi_12plus", "0|0|0|0|0|0|0|0|0|0")

    # Migration v12: Fix selected provider, since there is only one qbittorrent and one rtorrent provider
    def _migrate_v12(self):
        settings.TORRENT_METHOD = re.sub(r"^new_|9$", "", check_setting_str(settings.CFG, "General", "torrent_method", "blackhole"))



from sickchill.oldbeard import db


# Add new migrations at the bottom of the list; subclass the previous migration.
class InitialSchema(db.SchemaUpgrade):
    def test(self):
        return self.has_table("db_version")

    def execute(self):
        queries = (
            ("CREATE TABLE lastUpdate (provider TEXT, time NUMERIC);",),
            ("CREATE TABLE lastSearch (provider TEXT, time NUMERIC);",),
            (
                "CREATE TABLE scene_exceptions ("
                "exception_id INTEGER PRIMARY KEY, indexer_id INTEGER, show_name TEXT, season NUMERIC DEFAULT -1, custom NUMERIC DEFAULT 0);",
            ),
            ("CREATE TABLE scene_names (indexer_id INTEGER, name TEXT);",),
            ("CREATE TABLE network_timezones (network_name TEXT PRIMARY KEY, timezone TEXT);",),
            ("CREATE TABLE scene_exceptions_refresh (list TEXT PRIMARY KEY, last_refreshed INTEGER);",),
            ("CREATE TABLE db_version (db_version INTEGER);",),
            (
                "CREATE TABLE results (provider TEXT, name TEXT, season NUMERIC, episodes TEXT, indexerid NUMERIC, url TEXT, time NUMERIC, quality TEXT, "
                "release_group TEXT, version NUMERIC, seeders INTEGER DEFAULT 0, leechers INTEGER DEFAULT 0, size INTEGER DEFAULT -1, status INTEGER DEFAULT 0, "
                "failed INTEGER DEFAULT 0, added TEXT DEFAULT CURRENT_TIMESTAMP);",
            ),
            ("INSERT INTO db_version(db_version) VALUES (1);",),
            ("CREATE UNIQUE INDEX IF NOT EXISTS idx_url ON results (url);",),
            ("CREATE INDEX IF NOT EXISTS provider ON results (provider);",),
            ("CREATE INDEX IF NOT EXISTS seeders ON results (seeders);",),
        )
        for query in queries:
            if len(query) == 1:
                self.connection.action(query[0])
            else:
                self.connection.action(query[0], query[1:])


class ResultsTable(InitialSchema):
    def test(self):
        return self.has_table("results")

    def execute(self):
        import sickchill.settings

        self.connection.action(
            "CREATE TABLE results (provider TEXT, name TEXT, season NUMERIC, episodes TEXT, indexerid NUMERIC, url TEXT, time NUMERIC, quality TEXT,"
            "release_group TEXT, version NUMERIC, seeders INTEGER DEFAULT 0, leechers INTEGER DEFAULT 0, size INTEGER DEFAULT -1, status INTEGER DEFAULT 0, "
            "failed INTEGER DEFAULT 0, added TEXT DEFAULT CURRENT_TIMESTAMP);"
        )

        for provider in sickchill.settings.providerList:
            provider_id = provider.get_id()
            if self.has_table(provider_id):
                self.add_column(provider_id, "provider", col_type="TEXT", default=provider_id)
                self.add_column(provider_id, "seeders", col_type="INTEGER")
                self.add_column(provider_id, "leechers", col_type="INTEGER")
                self.add_column(provider_id, "size", col_type="INTEGER")
                self.add_column(provider_id, "status", col_type="INTEGER")
                self.add_column(provider_id, "failed", col_type="INTEGER")
                timestamp = self.connection.select("SELECT CURRENT_TIMESTAMP;")
                self.add_column(provider_id, "added", col_type="TEXT", default=timestamp[0]["CURRENT_TIMESTAMP"])

                # language=TEXT
                self.connection.action(
                    "INSERT INTO results SELECT provider, name , season, episodes, indexerid, url, time, quality,"
                    "release_group, version, seeders, leechers, size, status, failed, added FROM ?",
                    [provider_id],
                )
                self.connection.action("DROP TABLE {}".format(provider_id))



from abc import ABC

from sickchill import settings


class UpdateManagerBase(ABC):
    @staticmethod
    def get_update_url():
        return f"{settings.WEB_ROOT}/home/update/?pid={settings.PID}"

    def list_remote_branches(self):
        raise NotImplementedError

    def get_current_version(self):
        raise NotImplementedError

    def get_current_commit_hash(self):
        raise NotImplementedError

    def get_newest_version(self):
        raise NotImplementedError

    def get_newest_commit_hash(self):
        raise NotImplementedError

    def get_num_commits_behind(self):
        raise NotImplementedError

    def set_newest_text(self):
        raise NotImplementedError

    def need_update(self):
        raise NotImplementedError

    def update(self):
        raise NotImplementedError



from sickchill import logger, settings

from ..oldbeard import config, notifiers
from .common import PageTemplate
from .index import BaseHandler

login_error = ""


class LoginHandler(BaseHandler):
    def get(self, next_=None):
        next_ = self.get_query_argument("next", next_)
        if self.get_current_user():
            self.redirect(next_ or "/" + settings.DEFAULT_PAGE + "/")
        else:
            t = PageTemplate(rh=self, filename="login.mako")
            self.finish(t.render(title=_("Login"), header=_("Login"), topmenu="login", login_error=login_error))

    def post(self, next_=None):
        notifiers.notify_login(self.request.remote_ip)
        global login_error

        if self.get_body_argument("username", None) == settings.WEB_USERNAME and self.get_body_argument("password", None) == settings.WEB_PASSWORD:
            login_error = ""
            remember_me = config.checkbox_to_value(self.get_body_argument("remember_me", "0"))
            self.set_secure_cookie("sickchill_user", settings.API_KEY, expires_days=(None, 30)[remember_me])
            logger.info(_("User logged into the SickChill web interface"))
        else:
            logger.warning(_("User attempted a failed login to the SickChill web interface from IP: ") + self.request.remote_ip)
            login_error = _("Incorrect username or password! Both username and password are case sensitive!")

        next_ = self.get_query_argument("next", next_)
        self.redirect(next_ or "/" + settings.DEFAULT_PAGE + "/")


class LogoutHandler(BaseHandler):
    def get(self, next_=None):
        self.clear_cookie("sickchill_user")
        self.redirect("/login/")



import re
import traceback
from urllib.parse import urljoin

from requests.utils import dict_from_cookiejar

from sickchill import logger
from sickchill.helper.common import try_int
from sickchill.helper.exceptions import AuthException
from sickchill.oldbeard import tvcache
from sickchill.oldbeard.bs4_parser import BS4Parser
from sickchill.providers.torrent.TorrentProvider import TorrentProvider


class Provider(TorrentProvider):
    def __init__(self):

        # Provider Init
        super().__init__("Nebulance")

        # Credentials
        self.username = None
        self.password = None

        # Torrent Stats
        self.minseed = 0
        self.minleech = 0
        self.freeleech = None

        # URLs
        self.url = "https://nebulance.io/"
        self.urls = {
            "login": urljoin(self.url, "/login.php"),
            "search": urljoin(self.url, "/torrents.php"),
        }

        # Proper Strings

        # Cache
        self.cache = tvcache.TVCache(self)

    def _check_auth(self):

        if not self.username or not self.password:
            raise AuthException("Your authentication credentials for " + self.name + " are missing, check your config.")

        return True

    def login(self):
        if any(dict_from_cookiejar(self.session.cookies).values()):
            return True

        login_params = {"username": self.username, "password": self.password, "keeplogged": "on", "login": "Login"}

        response = self.get_url(self.urls["login"], post_data=login_params, returns="text")
        if not response:
            logger.warning("Unable to connect to provider")
            return False

        if re.search("Your username or password was incorrect.", response):
            logger.warning("Invalid username or password. Check your settings")
            return False

        return True

    def search(self, search_strings, age=0, ep_obj=None):
        results = []
        if not self.login():
            return results

        for mode in search_strings:
            items = []
            for search_string in {*search_strings[mode]}:

                if mode != "RSS":
                    logger.debug(_("Search String: {search_string}".format(search_string=search_string)))

                search_params = {
                    "searchtext": search_string,
                    "filter_freeleech": (0, 1)[self.freeleech is True],
                    "order_by": ("seeders", "time")[mode == "RSS"],
                    "order_way": "desc",
                }

                if not search_string:
                    del search_params["searchtext"]

                data = self.get_url(self.urls["search"], params=search_params, returns="text")
                if not data:
                    logger.debug("No data returned from provider")
                    continue

                try:
                    with BS4Parser(data, "html5lib") as html:
                        torrent_table = html.find("table", {"id": "torrent_table"})
                        if not torrent_table:
                            logger.debug("Data returned from {0} does not contain any torrents".format(self.name))
                            continue

                        labels = [x.get_text(strip=True) or x.a.img.get("alt") for x in torrent_table.find("tr", class_="colhead").find_all("td")]
                        torrent_rows = torrent_table("tr", class_="torrent")

                        # Continue only if one Release is found
                        if not torrent_rows:
                            logger.debug("Data returned from {0} does not contain any torrents".format(self.name))
                            continue

                        for torrent_row in torrent_rows:
                            freeleech = torrent_row.find("img", alt="Freeleech") is not None
                            if self.freeleech and not freeleech:
                                continue

                            # Normal Download Link
                            download_item = torrent_row.find("a", {"title": "Download Torrent"})

                            if not download_item:
                                # If the user has downloaded it
                                download_item = torrent_row.find("a", {"title": "Previously Grabbed Torrent File"})
                            if not download_item:
                                # If the user is seeding
                                download_item = torrent_row.find("a", {"title": "Currently Seeding Torrent"})
                            if not download_item:
                                # If the user is leeching
                                download_item = torrent_row.find("a", {"title": "Currently Leeching Torrent"})
                            if not download_item:
                                # If there are none
                                continue

                            download_url = urljoin(self.url, download_item["href"])

                            temp_anchor = torrent_row.find("a", {"data-src": True})
                            title = temp_anchor["data-src"].rsplit(".", 1)[0]
                            if not all([title, download_url]):
                                continue

                            cells = torrent_row("td")
                            seeders = try_int(cells[labels.index("∧")].text.strip())
                            leechers = try_int(cells[labels.index("∨")].get_text(strip=True))

                            # Filter unseeded torrent
                            if seeders < self.minseed or leechers < self.minleech:
                                if mode != "RSS":
                                    logger.debug(
                                        "Discarding torrent because it doesn't meet the"
                                        " minimum seeders or leechers: {0} (S:{1} L:{2})".format(title, seeders, leechers)
                                    )
                                continue

                            size = temp_anchor["data-filesize"] or -1

                            item = {"title": title, "link": download_url, "size": size, "seeders": seeders, "leechers": leechers, "hash": ""}
                            if mode != "RSS":
                                logger.debug("Found result: {0} with {1} seeders and {2} leechers".format(title, seeders, leechers))

                            items.append(item)
                except Exception:
                    logger.exception("Failed parsing provider. Traceback: {0}".format(traceback.format_exc()))

            # For each search mode sort all the items by seeders
            items.sort(key=lambda d: try_int(d.get("seeders", 0)), reverse=True)
            results += items

        return results



#!/usr/bin/env python3

import datetime
import os
import platform
import shutil
import signal
import subprocess
import sys
import threading
import time
import traceback

import sickchill.start
from sickchill import logger, settings
from sickchill.init_helpers import check_installed, remove_pid_file, setup_gettext
from sickchill.movies import MovieList

setup_gettext()

import mimetypes
from pathlib import Path

mimetypes.add_type("text/css", ".css")
mimetypes.add_type("application/sfont", ".otf")
mimetypes.add_type("application/sfont", ".ttf")
mimetypes.add_type("application/javascript", ".js")
mimetypes.add_type("application/font-woff", ".woff")
# Not sure about this one, but we also have halflings in .woff so I think it wont matter
# mimetypes.add_type("application/font-woff2", ".woff2")

from configobj import ConfigObj

from sickchill.helper.argument_parser import SickChillArgumentParser
from sickchill.oldbeard import db, failed_history, name_cache, network_timezones
from sickchill.oldbeard.event_queue import Events
from sickchill.tv import TVShow
from sickchill.update_manager import GitUpdateManager, SourceUpdateManager
from sickchill.views.server_settings import SRWebServer

# http://bugs.python.org/issue7980#msg221094
THROWAWAY = datetime.datetime.strptime("20110101", "%Y%m%d")

signal.signal(signal.SIGINT, sickchill.start.sig_handler)
signal.signal(signal.SIGTERM, sickchill.start.sig_handler)


class SickChill(object):

    """
    Main SickChill module
    """

    def __init__(self):
        # system event callback for shutdown/restart
        settings.events = Events(self.shutdown)

        # daemon constants
        self.run_as_daemon = False

        # web server constants
        self.web_server = None
        self.forced_port = None
        self.no_launch = False

        self.start_port = settings.WEB_PORT

        self.console_logging = True

    @staticmethod
    def clear_cache():
        """
        Remove the Mako cache directory
        """
        try:
            cache_folder = os.path.join(settings.CACHE_DIR, "mako")
            if os.path.isdir(cache_folder):
                shutil.rmtree(cache_folder)
        except Exception:
            logger.warning("Unable to remove the cache/mako directory!")

    def start(self):
        """
        Start SickChill
        """
        # do some preliminary stuff
        settings.MY_FULLNAME = os.path.normpath(os.path.abspath(__file__))
        settings.MY_NAME = os.path.basename(settings.MY_FULLNAME)

        settings.DATA_DIR = os.path.dirname(settings.PROG_DIR)
        profile_path = str(Path.home().joinpath("sickchill").absolute())
        if check_installed():
            settings.DATA_DIR = profile_path

        if settings.DATA_DIR != profile_path:
            checks = ["sickbeard.db", "sickchill.db", "config.ini"]
            if not any([os.path.isfile(os.path.join(settings.DATA_DIR, check)) for check in checks]):
                settings.DATA_DIR = profile_path

        settings.MY_ARGS = sys.argv[1:]

        # Rename the main thread
        threading.currentThread().name = "MAIN"

        args = SickChillArgumentParser(settings.DATA_DIR).parse_args()

        if args.force_update:
            result = self.force_update()
            sys.exit(int(not result))  # Ok -> 0 , Error -> 1

        # Need console logging for SickChill.py and SickChill-console.exe
        settings.NO_RESIZE = args.noresize
        self.console_logging = not (hasattr(sys, "frozen") or args.quiet or args.daemon)
        self.no_launch = args.nolaunch or args.daemon
        self.forced_port = args.port
        self.run_as_daemon = args.daemon and platform.system() != "Windows"

        # The pid file is only useful in daemon mode, make sure we can write the file properly
        if bool(args.pidfile) and not self.run_as_daemon:
            if self.console_logging:
                sys.stdout.write("Not running in daemon mode. PID file creation disabled.\n")

        settings.DATA_DIR = os.path.abspath(args.datadir) if args.datadir else settings.DATA_DIR
        settings.CONFIG_FILE = os.path.abspath(args.config) if args.config else os.path.join(settings.DATA_DIR, "config.ini")

        # Make sure that we can create the data dir
        if not os.access(settings.DATA_DIR, os.F_OK):
            try:
                os.makedirs(settings.DATA_DIR, 0o744)
            except os.error:
                raise SystemExit("Unable to create data directory: {0}".format(settings.DATA_DIR))

        # Make sure we can write to the data dir
        if not os.access(settings.DATA_DIR, os.W_OK):
            raise SystemExit("Data directory must be writeable: {0}".format(settings.DATA_DIR))

        # Make sure we can write to the config file
        if not os.access(settings.CONFIG_FILE, os.W_OK):
            if os.path.isfile(settings.CONFIG_FILE):
                raise SystemExit("Config file must be writeable: {0}".format(settings.CONFIG_FILE))
            elif not os.access(os.path.dirname(settings.CONFIG_FILE), os.W_OK):
                raise SystemExit("Config file root dir must be writeable: {0}".format(os.path.dirname(settings.CONFIG_FILE)))

        os.chdir(settings.DATA_DIR)

        # Check if we need to perform a restore first
        restore_dir = os.path.join(settings.DATA_DIR, "restore")
        if os.path.exists(restore_dir):
            success = self.restore_db(restore_dir, settings.DATA_DIR)
            if self.console_logging:
                sys.stdout.write("Restore: restoring DB and config.ini {0}!\n".format(("FAILED", "SUCCESSFUL")[success]))

        # Load the config and publish it to the oldbeard package
        if self.console_logging and not os.path.isfile(settings.CONFIG_FILE):
            sys.stdout.write("Unable to find {0}, all settings will be default!\n".format(settings.CONFIG_FILE))

        settings.CFG = ConfigObj(settings.CONFIG_FILE, encoding="UTF-8", indent_type="  ")

        # Initialize the config and our threads
        sickchill.start.initialize(consoleLogging=self.console_logging)

        # Get PID
        settings.PID = os.getpid()

        # Build from the DB to start with
        self.load_shows_from_db()

        logger.info("Starting SickChill [{branch}] using '{config}'".format(branch=settings.BRANCH, config=settings.CONFIG_FILE))

        self.clear_cache()

        if settings.DEVELOPER:
            settings.movie_list = MovieList()

        web_options = {}
        if self.forced_port:
            logger.info("Forcing web server to port {port}".format(port=self.forced_port))
            self.start_port = self.forced_port
            web_options.update(
                {
                    "port": int(self.start_port),
                }
            )
        else:
            self.start_port = settings.WEB_PORT

        # start web server
        self.web_server = SRWebServer(web_options)
        self.web_server.start()

        # Fire up all our threads
        sickchill.start.start()

        # Build internal name cache
        name_cache.build_name_cache()

        # Pre-populate network timezones, it isn't thread safe
        network_timezones.update_network_dict()

        # sure, why not?
        if settings.USE_FAILED_DOWNLOADS:
            failed_history.trimHistory()

        # Check for metadata indexer updates for shows (sets the next aired ep!)
        # oldbeard.showUpdateScheduler.forceRun()

        # Launch browser
        if settings.LAUNCH_BROWSER and not self.no_launch:
            sickchill.start.launchBrowser("https" if settings.ENABLE_HTTPS else "http", self.start_port, settings.WEB_ROOT)

        # main loop
        while True:
            time.sleep(1)

    @staticmethod
    def load_shows_from_db():
        """
        Populates the showList with shows from the database
        """
        logger.debug("Loading initial show list")

        main_db_con = db.DBConnection()
        sql_results = main_db_con.select("SELECT indexer, indexer_id, location FROM tv_shows;")

        settings.showList = []
        for sql_show in sql_results:
            try:
                cur_show = TVShow(sql_show["indexer"], sql_show["indexer_id"])
                cur_show.nextEpisode()
                settings.showList.append(cur_show)
            except Exception as error:
                logger.exception("There was an error creating the show in {0}: Error {1}".format(sql_show["location"], error))
                logger.debug(traceback.format_exc())

    @staticmethod
    def restore_db(src_dir, dst_dir):
        """
        Restore the Database from a backup

        :param src_dir: Directory containing backup
        :param dst_dir: Directory to restore to
        :return:
        """
        try:
            files_list = ["sickbeard.db", "sickchill.db", "config.ini", "failed.db", "cache.db"]
            for filename in files_list:
                src_file = os.path.join(src_dir, filename)
                dst_file = os.path.join(dst_dir, filename)
                bak_file = os.path.join(dst_dir, "{0}.bak-{1}".format(filename, datetime.datetime.now().strftime("%Y%m%d_%H%M%S")))
                sickchill_db = os.path.join(dst_dir, "sickchill.db")
                sickbeard_db = os.path.join(src_dir, "sickbeard.db")
                if os.path.isfile(src_file):
                    if src_file == sickbeard_db:
                        dst_file = sickchill_db

                    if os.path.isfile(dst_file):
                        shutil.move(dst_file, bak_file)
                    shutil.move(src_file, dst_file)

            sickbeard_db = os.path.join(dst_dir, "sickbeard.db")
            sickchill_db = os.path.join(dst_dir, "sickchill.db")
            if os.path.isfile(sickbeard_db) and not os.path.isfile(sickchill_db):
                shutil.move(sickbeard_db, sickchill_db)

            return True
        except Exception:
            return False

    def shutdown(self, event):
        """
        Shut down SickChill

        :param event: Type of shutdown event, used to see if restart required
        """
        if settings.started:
            sickchill.start.halt()  # stop all tasks
            sickchill.start.saveAll()  # save all shows to DB

            # shutdown web server
            if self.web_server:
                logger.info("Shutting down Tornado")
                self.web_server.shutdown()

                try:
                    self.web_server.join(10)
                except Exception:
                    pass

            self.clear_cache()  # Clean cache

            # if run as daemon delete the pid file
            remove_pid_file()

            if event == sickchill.oldbeard.event_queue.Events.SystemEvent.RESTART:
                popen_list = [sys.executable, settings.MY_FULLNAME]
                if popen_list and not settings.NO_RESTART:
                    popen_list += settings.MY_ARGS
                    if "--nolaunch" not in popen_list:
                        popen_list += ["--nolaunch"]
                    logger.info("Restarting SickChill with {options}".format(options=popen_list))
                    # shutdown the logger to make sure it's released the logfile BEFORE it restarts SC.
                    logger.shutdown()
                    subprocess.Popen(popen_list, cwd=os.getcwd(), universal_newlines=True)

        # Make sure the logger has stopped, just in case
        logger.shutdown()
        os._exit(0)

    @staticmethod
    def force_update():
        """
        Forces SickChill to update to the latest version and exit.

        :return: True if successful, False otherwise
        """

        def update_with_git():
            def run_git(updater, cmd):
                stdout_, stderr_, exit_status = updater._run_git(updater._git_path, cmd)
                if not exit_status == 0:
                    print("Failed to run command: {0} {1}".format(updater._git_path, cmd))
                    return False
                else:
                    return True

            updater = GitUpdateManager()
            if not run_git(updater, "config remote.origin.url https://github.com/SickChill/SickChill.git"):
                return False
            if not run_git(updater, "fetch origin --prune"):
                return False
            if not run_git(updater, "checkout master"):
                return False
            if not run_git(updater, "reset --hard origin/master"):
                return False

            return True

        if os.path.isdir(os.path.join(os.path.dirname(settings.PROG_DIR), ".git")):  # update with git
            print("Forcing SickChill to update using git...")
            result = update_with_git()
            if result:
                print("Successfully updated to latest commit. You may now run SickChill normally.")
                return True
            else:
                print("Error while trying to force an update using git.")

        print("Forcing SickChill to update using source...")
        if not SourceUpdateManager().update():
            print("Failed to force an update.")
            return False

        print("Successfully updated to latest commit. You may now run SickChill normally.")
        return True


if __name__ == "__main__":
    # start SickChill
    SickChill().start()
    remove_pid_file()



import locale
import logging
import logging.handlers
import os
import platform
import re
import sys
import traceback
from logging import NullHandler
from urllib.parse import quote

from github import InputFileContent
from github.GithubException import RateLimitExceededException, TwoFactorException

from sickchill import settings
from sickchill.helper.common import dateTimeFormat
from sickchill.oldbeard import classes

# log levels
ERROR = logging.ERROR
WARNING = logging.WARNING
INFO = logging.INFO
DEBUG = logging.DEBUG
DB = 5

LOGGING_LEVELS = {
    "ERROR": ERROR,
    "WARNING": WARNING,
    "INFO": INFO,
    "DEBUG": DEBUG,
    "DB": DB,
}

censored_items = {}


class DispatchFormatter(logging.Formatter, object):
    """
    Censor information such as API keys, user names, and passwords from the Log
    """

    def __init__(self, fmt=None, datefmt=None, style="{"):
        super().__init__(fmt, datefmt, style=style)

    def format(self, record):
        """
        Strips censored items from string and formats the log line

        :param record: to format
        """

        msg = record.msg

        # set of censored items
        censored = {item for _, item in censored_items.items() if item}
        # set of censored items and urlencoded counterparts
        censored = list(censored | {quote(item) for item in censored})
        # sort the list in order of descending length so that entire item is censored
        # e.g. password and password_1 both get censored instead of getting ********_1
        censored.sort(key=len, reverse=True)

        if not isinstance(msg, (str, bytes)):
            msg = repr(msg)

        for item in censored:
            try:
                # passwords that include ++ for example will error. Cannot escape or it wont match at all.
                msg = re.sub(fr"\b({item})\b", "*" * 8, msg)
            except re.error:
                msg = msg.replace(item, "*" * 8)
            except TypeError:
                print(msg)

        # Needed because Newznab apikey isn't stored as key=value in a section.
        msg = re.sub(r"([&?]r|[&?]apikey|[&?]jackett_apikey|[&?]api_key)(?:=|%3D)[^&]*([&\w]?)", r"\1=**********\2", msg, re.I)

        if record.levelno == ERROR:
            classes.ErrorViewer.add(classes.UIError(msg))
        elif record.levelno == WARNING:
            classes.WarningViewer.add(classes.UIError(msg))

        return super().format(record)


class Logger(object):
    """
    Logger to create log entries
    """

    def __init__(self):
        self.logger = logging.getLogger("sickchill")

        self.loggers = [
            logging.getLogger("sickchill"),
            logging.getLogger("sickchill.movie"),
            logging.getLogger("tornado.general"),
            logging.getLogger("tornado.application"),
            logging.getLogger("subliminal"),
            logging.getLogger("tornado.access"),
            logging.getLogger("imdbpy.parser.http.piculet"),
        ]

        self.console_logging = False
        self.file_logging = False
        self.debug_logging = False
        self.database_logging = False
        self.log_file = None

        self.submitter_running = False

    def init_logging(self, console_logging=False, file_logging=False, debug_logging=False, database_logging=False):
        """
        Initialize logging

        :param console_logging: True if logging to console
        :param file_logging: True if logging to file
        :param debug_logging: True if debug logging is enabled
        :param database_logging: True if logging database access
        """
        self.log_file = self.log_file or os.path.join(settings.LOG_DIR, "sickchill.log")

        global log_file
        log_file = self.log_file

        self.debug_logging = debug_logging
        self.console_logging = console_logging
        self.file_logging = file_logging
        self.database_logging = database_logging

        logging.addLevelName(DB, "DB")  # add a new logging level DB
        logging.getLogger().addHandler(NullHandler())  # nullify root logger

        # set custom root logger
        for logger in self.loggers:
            if logger is not self.logger:
                logger.root = self.logger
                logger.parent = self.logger

        log_level = DB if self.database_logging else DEBUG if self.debug_logging else INFO

        # set minimum logging level allowed for loggers
        for logger in self.loggers:
            if logger.name in ("subliminal", "tornado.access", "tornado.general", "imdbpy.parser.http.piculet"):
                logger.setLevel("CRITICAL")
            else:
                logger.setLevel(log_level)

        log_format = "{asctime} {levelname} :: {threadName} :: {message}"
        # console log handler
        if self.console_logging:
            console = logging.StreamHandler()
            console.setFormatter(DispatchFormatter(log_format, dateTimeFormat))
            console.setLevel(log_level)

            for logger in self.loggers:
                logger.addHandler(console)

        # rotating log file handler
        if self.file_logging:
            rfh = logging.handlers.RotatingFileHandler(self.log_file, maxBytes=int(settings.LOG_SIZE * 1048576), backupCount=settings.LOG_NR)
            rfh.setFormatter(DispatchFormatter(log_format, dateTimeFormat))
            rfh.setLevel(log_level)

            for logger in self.loggers:
                logger.addHandler(rfh)

    def set_level(self):
        """
        Sets the logging level for root and all of our loggers
        """
        self.debug_logging = settings.DEBUG
        self.database_logging = settings.DBDEBUG

        level = DB if self.database_logging else DEBUG if self.debug_logging else INFO
        for logger in self.loggers:
            if logger.name in ("subliminal", "tornado.access", "tornado.general"):
                logger.setLevel("CRITICAL")
            else:
                logger.setLevel(level)
                for handler in logger.handlers:
                    handler.setLevel(level)

    @staticmethod
    def shutdown():
        """
        Shut down the logger
        """
        logging.shutdown()

    def log_error_and_exit(self, error_msg, *args, **kwargs):
        self.logger.log(ERROR, error_msg, *args, **kwargs)

        if not self.console_logging:
            sys.exit(error_msg)
        else:
            sys.exit(1)

    def submit_errors(self):

        submitter_result = ""
        issue_id = None

        if not all((settings.GIT_TOKEN, settings.DEBUG, settings.gh, classes.ErrorViewer.errors)):
            submitter_result = "Please set your GitHub token in the config and enable debug. Unable to submit issue ticket to GitHub!"
            return submitter_result, issue_id

        try:
            from sickchill.update_manager import UpdateManager

            update_manager = UpdateManager()
            update_manager.check_for_new_version()
            commits_behind = update_manager.get_num_commits_behind()
        except Exception:
            submitter_result = "Could not check if your SickChill is updated, unable to submit issue ticket to GitHub!"
            return submitter_result, issue_id

        if commits_behind is None or commits_behind > 0:
            submitter_result = "Please update SickChill, unable to submit issue ticket to GitHub with an outdated version!"
            return submitter_result, issue_id

        if self.submitter_running:
            submitter_result = "Issue submitter is running, please wait for it to complete"
            return submitter_result, issue_id

        self.submitter_running = True

        try:
            # read log file
            __log_data = None

            if os.path.isfile(self.log_file):
                with open(self.log_file) as log_f:
                    __log_data = log_f.readlines()

            for i in range(1, int(settings.LOG_NR)):
                f_name = f"{self.log_file}.{i}"
                if os.path.isfile(f_name) and (len(__log_data) <= 500):
                    with open(f_name) as log_f:
                        __log_data += log_f.readlines()

            __log_data = list(reversed(__log_data))

            # parse and submit errors to issue tracker
            for cur_error in sorted(classes.ErrorViewer.errors, key=lambda error: error.time, reverse=True)[:500]:
                try:
                    title_error = str(cur_error.title)
                    if not title_error or title_error == "None":
                        title_error = re.match(r"^[A-Za-z0-9\-\[\] :]+::\s(?:\[[\w]{7}\])\s*(.*)$", cur_error.message).group(1)

                    if len(title_error) > 1000:
                        title_error = title_error[0:1000]

                except Exception as err_msg:
                    self.logger.log(ERROR, f"Unable to get error title : {err_msg}")
                    title_error = "UNKNOWN"

                gist = None
                regex = fr"^(?P<time>{re.escape(cur_error.time)})\s+(?P<level>[A-Z]+)\s+[A-Za-z0-9\-\[\] :]+::.*$"
                for i, data in enumerate(__log_data):
                    match = re.match(regex, data)
                    if match:
                        level = match.group("level")
                        if LOGGING_LEVELS[level] == ERROR:
                            paste_data = "".join(__log_data[i : i + 50])
                            if paste_data:
                                gist = settings.gh.get_user().create_gist(False, {"sickchill.log": InputFileContent(paste_data)})
                            break
                    else:
                        gist = "No ERROR found"

                try:
                    locale_name = locale.getdefaultlocale()[1]
                except Exception:
                    locale_name = "unknown"

                if gist and gist != "No ERROR found":
                    log_link = f"Link to Log: {gist.html_url}"
                else:
                    log_link = "No Log available with ERRORS:"

                msg = [
                    "### INFO",
                    f"Python Version: **{sys.version[:120]}**".replace("\n", ""),
                    f"Operating System: **{platform.platform()}**",
                    f"Locale: {locale_name}",
                    f"Branch: **{settings.BRANCH}**",
                    f"Commit: SickChill/SickChill@{settings.CUR_COMMIT_HASH}",
                    log_link,
                    "### ERROR",
                    "```",
                    cur_error.message,
                    "```",
                    "---",
                    "_STAFF NOTIFIED_: @SickChill/owners @SickChill/moderators",
                ]

                message = "\n".join(msg)
                title_error = f"[APP SUBMITTED]: {title_error}"

                repo = settings.gh.get_organization(settings.GIT_ORG).get_repo(settings.GIT_REPO)
                reports = repo.get_issues(state="all")

                def is_ascii_error(title):
                    # [APP SUBMITTED]: 'ascii' codec can't encode characters in position 00-00: ordinal not in range(128)
                    # [APP SUBMITTED]: 'charmap' codec can't decode byte 0x00 in position 00: character maps to <undefined>
                    return re.search(r".* codec can\'t .*code .* in position .*:", title) is not None

                def is_malformed_error(title):
                    # [APP SUBMITTED]: not well-formed (invalid token): line 0, column 0
                    return re.search(r".* not well-formed \(invalid token\): line .* column .*", title) is not None

                ascii_error = is_ascii_error(title_error)
                malformed_error = is_malformed_error(title_error)

                issue_found = False
                for report in reports:
                    if (
                        title_error.rsplit(" :: ")[-1] in report.title
                        or (malformed_error and is_malformed_error(report.title))
                        or (ascii_error and is_ascii_error(report.title))
                    ):

                        issue_id = report.number
                        if not report.raw_data["locked"]:
                            if report.create_comment(message):
                                submitter_result = f"Commented on existing issue #{issue_id} successfully!"
                            else:
                                submitter_result = f"Failed to comment on found issue #{issue_id}!"
                        else:
                            submitter_result = f"Issue #{issue_id} is locked, check GitHub to find info about the error."

                        issue_found = True
                        break

                if not issue_found:
                    issue = repo.create_issue(title_error, message)
                    if issue:
                        issue_id = issue.number
                        submitter_result = f"Your issue ticket #{issue_id} was submitted successfully!"
                    else:
                        submitter_result = "Failed to create a new issue!"

                if issue_id and cur_error in classes.ErrorViewer.errors:
                    # clear error from error list
                    classes.ErrorViewer.errors.remove(cur_error)
        except RateLimitExceededException:
            submitter_result = "Your Github user has exceeded its API rate limit, please try again later"
            issue_id = None
        except TwoFactorException:
            submitter_result = "Your Github account requires Two-Factor Authentication, " "please change your auth method in the config"
            issue_id = None
        except Exception:
            self.logger.log(ERROR, traceback.format_exc())
            submitter_result = "Exception generated in issue submitter, please check the log"
            issue_id = None
        finally:
            self.submitter_running = False

        return submitter_result, issue_id


class Wrapper(object):
    instance = Logger()

    def __init__(self, wrapped):
        self.wrapped = wrapped

    def __getattr__(self, name):
        try:
            return getattr(self.wrapped, name)
        except AttributeError:
            return getattr(self.instance, name)


_globals = sys.modules[__name__] = Wrapper(sys.modules[__name__])

LOG_FILTERS = {
    "<NONE>": _("&lt;No Filter&gt;"),
    "DAILYSEARCHER": _("Daily Searcher"),
    "BACKLOG": _("Backlog"),
    "SHOWUPDATER": _("Show Updater"),
    "CHECKVERSION": _("Check Version"),
    "SHOWQUEUE": _("Show Queue"),
    "SEARCHQUEUE": _("Search Queue (All)"),
    "SEARCHQUEUE-DAILY-SEARCH": _("Search Queue (Daily Searcher)"),
    "SEARCHQUEUE-BACKLOG": _("Search Queue (Backlog)"),
    "SEARCHQUEUE-MANUAL": _("Search Queue (Manual)"),
    "SEARCHQUEUE-RETRY": _("Search Queue (Retry/Failed)"),
    "SEARCHQUEUE-RSS": _("Search Queue (RSS)"),
    "FINDPROPERS": _("Find Propers"),
    "POSTPROCESSOR": _("Postprocessor"),
    "FINDSUBTITLES": _("Find Subtitles"),
    "TRAKTCHECKER": _("Trakt Checker"),
    "EVENT": _("Event"),
    "ERROR": _("Error"),
    "TORNADO": _("Tornado"),
    "Thread": _("Thread"),
    "MAIN": _("Main"),
}


def log_data(min_level, log_filter, log_search, max_lines):
    regex = r"^\d{4}\-\d{2}\-\d{2} \d{2}\:\d{2}:\d{2} ([A-Z]+) \:\: (.+?) \:\: (.*)$"
    if log_filter not in LOG_FILTERS:
        log_filter = "<NONE>"

    final_data = []

    log_files = []
    if os.path.isfile(Wrapper.instance.log_file):
        log_files.append(Wrapper.instance.log_file)

        for i in range(1, int(settings.LOG_NR)):
            name = Wrapper.instance.log_file + "." + str(i)
            if not os.path.isfile(name):
                break
            log_files.append(name)
    else:
        return final_data

    data = []
    for _log_file in log_files:
        if len(data) < max_lines:
            with open(_log_file, "r") as f:
                data += [line.strip() + "\n" for line in reversed(f.readlines()) if line.strip()]
        else:
            break

    found_lines = 0
    for x in data:
        match = re.match(regex, x)

        if match:
            level = match.group(1)
            log_name = match.group(2)

            if not settings.DEBUG and level == "DEBUG":
                continue

            if not settings.DBDEBUG and level == "DB":
                continue

            if level not in LOGGING_LEVELS:
                final_data.append("AA " + x)
                found_lines += 1
            elif log_search and log_search.lower() in x.lower():
                final_data.append(x)
                found_lines += 1
            elif not log_search and LOGGING_LEVELS[level] >= int(min_level) and (log_filter == "<NONE>" or log_name.startswith(log_filter)):
                final_data.append(x)
                found_lines += 1
        else:
            final_data.append("AA " + x)
            found_lines += 1

        if found_lines >= max_lines:
            break

    return final_data


log_error_and_exit = Wrapper.instance.log_error_and_exit
log_file = Wrapper.instance.log_file

log = Wrapper.instance.logger.log
debug = Wrapper.instance.logger.debug
info = Wrapper.instance.logger.info
warning = Wrapper.instance.logger.warning
error = Wrapper.instance.logger.error
exception = error
critical = Wrapper.instance.logger.critical


def database(msg, *args, **kwargs):
    """
    Handler to add database sql logging
    @param msg: message to be logged
    @param args: arguments to logger
    @param kwargs:  kwargs to logger
    """
    if settings.DBDEBUG:
        debug(msg, args, kwargs)


submit_errors = Wrapper.instance.submit_errors
init_logging = Wrapper.instance.init_logging
set_level = Wrapper.instance.set_level
shutdown = Wrapper.instance.shutdown



import os

from tornado.web import addslash

import sickchill.start
from sickchill import logger, settings
from sickchill.helper import try_int
from sickchill.oldbeard import config, naming, ui
from sickchill.oldbeard.common import NAMING_LIMITED_EXTEND_E_PREFIXED
from sickchill.views.common import PageTemplate
from sickchill.views.routes import Route

from . import Config


@Route("/config/postProcessing(/?.*)", name="config:postprocessing")
class ConfigPostProcessing(Config):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @addslash
    def index(self, *args_, **kwargs_):
        t = PageTemplate(rh=self, filename="config_postProcessing.mako")

        return t.render(
            submenu=self.ConfigMenu(),
            title=_("Config - Post Processing"),
            header=_("Post Processing"),
            topmenu="config",
            controller="config",
            action="postProcessing",
        )

    def savePostProcessing(
        self,
        kodi_data=None,
        mediabrowser_data=None,
        sony_ps3_data=None,
        wdtv_data=None,
        tivo_data=None,
        mede8er_data=None,
        keep_processed_dir=None,
        process_method=None,
        processor_follow_symlinks=None,
        del_rar_contents=None,
        process_automatically=None,
        no_delete=None,
        rename_episodes=None,
        airdate_episodes=None,
        file_timestamp_timezone=None,
        unpack=None,
        unpack_dir=None,
        unrar_tool=None,
        unar_tool=None,
        bsdtar_tool=None,
        move_associated_files=None,
        delete_non_associated_files=None,
        sync_files=None,
        postpone_if_sync_files=None,
        allowed_extensions=None,
        tv_download_dir=None,
        create_missing_show_dirs=None,
        add_shows_wo_dir=None,
        extra_scripts=None,
        nfo_rename=None,
        naming_pattern=None,
        naming_multi_ep=None,
        naming_custom_abd=None,
        naming_anime=None,
        naming_abd_pattern=None,
        naming_strip_year=None,
        naming_custom_sports=None,
        naming_sports_pattern=None,
        naming_custom_anime=None,
        naming_anime_pattern=None,
        naming_anime_multi_ep=None,
        autopostprocessor_frequency=None,
        use_icacls=None,
    ):

        results = []

        if not config.change_tv_download_dir(tv_download_dir):
            results += ["Unable to create directory " + os.path.normpath(tv_download_dir) + ", dir not changed."]

        config.change_postprocessor_frequency(autopostprocessor_frequency)
        config.change_process_automatically(process_automatically)
        settings.USE_ICACLS = config.checkbox_to_value(use_icacls)

        config.change_unrar_tool(unrar_tool, unar_tool, bsdtar_tool)

        unpack = try_int(unpack)
        if unpack == settings.UNPACK_PROCESS_CONTENTS:
            settings.UNPACK = int(self.isRarSupported() != "not supported")
            if settings.UNPACK != settings.UNPACK_PROCESS_CONTENTS:
                results.append(_("Unpacking Not Supported, disabling unpack setting"))
        elif unpack in settings.unpackStrings:
            settings.UNPACK = unpack

        if not config.change_unpack_dir(unpack_dir):
            results += ["Unable to change unpack directory to " + os.path.normpath(unpack_dir) + ", check the logs."]

        settings.NO_DELETE = config.checkbox_to_value(no_delete)
        settings.KEEP_PROCESSED_DIR = config.checkbox_to_value(keep_processed_dir)
        settings.CREATE_MISSING_SHOW_DIRS = config.checkbox_to_value(create_missing_show_dirs)
        settings.ADD_SHOWS_WO_DIR = config.checkbox_to_value(add_shows_wo_dir)
        settings.PROCESS_METHOD = process_method
        settings.PROCESSOR_FOLLOW_SYMLINKS = config.checkbox_to_value(processor_follow_symlinks)
        settings.DELRARCONTENTS = config.checkbox_to_value(del_rar_contents)
        settings.EXTRA_SCRIPTS = [x.strip() for x in extra_scripts.split("|") if x.strip()]
        settings.RENAME_EPISODES = config.checkbox_to_value(rename_episodes)
        settings.AIRDATE_EPISODES = config.checkbox_to_value(airdate_episodes)
        settings.FILE_TIMESTAMP_TIMEZONE = file_timestamp_timezone
        settings.MOVE_ASSOCIATED_FILES = config.checkbox_to_value(move_associated_files)
        settings.DELETE_NON_ASSOCIATED_FILES = config.checkbox_to_value(delete_non_associated_files)
        settings.SYNC_FILES = sync_files
        settings.POSTPONE_IF_SYNC_FILES = config.checkbox_to_value(postpone_if_sync_files)

        settings.ALLOWED_EXTENSIONS = ",".join({x.strip() for x in allowed_extensions.split(",") if x.strip()})
        settings.NAMING_CUSTOM_ABD = config.checkbox_to_value(naming_custom_abd)
        settings.NAMING_CUSTOM_SPORTS = config.checkbox_to_value(naming_custom_sports)
        settings.NAMING_CUSTOM_ANIME = config.checkbox_to_value(naming_custom_anime)
        settings.NAMING_STRIP_YEAR = config.checkbox_to_value(naming_strip_year)
        settings.NFO_RENAME = config.checkbox_to_value(nfo_rename)

        settings.METADATA_KODI = kodi_data
        settings.METADATA_MEDIABROWSER = mediabrowser_data
        settings.METADATA_PS3 = sony_ps3_data
        settings.METADATA_WDTV = wdtv_data
        settings.METADATA_TIVO = tivo_data
        settings.METADATA_MEDE8ER = mede8er_data

        settings.metadata_provider_dict["KODI"].set_config(settings.METADATA_KODI)
        settings.metadata_provider_dict["MediaBrowser"].set_config(settings.METADATA_MEDIABROWSER)
        settings.metadata_provider_dict["Sony PS3"].set_config(settings.METADATA_PS3)
        settings.metadata_provider_dict["WDTV"].set_config(settings.METADATA_WDTV)
        settings.metadata_provider_dict["TIVO"].set_config(settings.METADATA_TIVO)
        settings.metadata_provider_dict["Mede8er"].set_config(settings.METADATA_MEDE8ER)

        if self.isNamingValid(naming_pattern, naming_multi_ep) != "invalid":
            settings.NAMING_PATTERN = naming_pattern
            settings.NAMING_MULTI_EP = try_int(naming_multi_ep, NAMING_LIMITED_EXTEND_E_PREFIXED)
            settings.NAMING_FORCE_FOLDERS = naming.check_force_season_folders()
        else:
            results.append(_("You tried saving an invalid normal naming config, not saving your naming settings"))

        if self.isNamingValid(naming_anime_pattern, naming_anime_multi_ep, anime_type=naming_anime) != "invalid":
            settings.NAMING_ANIME_PATTERN = naming_anime_pattern
            settings.NAMING_ANIME_MULTI_EP = try_int(naming_anime_multi_ep, NAMING_LIMITED_EXTEND_E_PREFIXED)
            settings.NAMING_ANIME = try_int(naming_anime, 3)
            settings.NAMING_FORCE_FOLDERS = naming.check_force_season_folders()
        else:
            results.append(_("You tried saving an invalid anime naming config, not saving your naming settings"))

        if self.isNamingValid(naming_abd_pattern, None, abd=True) != "invalid":
            settings.NAMING_ABD_PATTERN = naming_abd_pattern
        else:
            results.append("You tried saving an invalid air-by-date naming config, not saving your air-by-date settings")

        if self.isNamingValid(naming_sports_pattern, None, sports=True) != "invalid":
            settings.NAMING_SPORTS_PATTERN = naming_sports_pattern
        else:
            results.append("You tried saving an invalid sports naming config, not saving your sports settings")

        sickchill.start.save_config()

        if results:
            for x in results:
                logger.warning(x)
            ui.notifications.error(_("Error(s) Saving Configuration"), "<br>\n".join(results))
        else:
            ui.notifications.message(_("Configuration Saved"), os.path.join(settings.CONFIG_FILE))

        return self.redirect("/config/postProcessing/")

    @staticmethod
    def testNaming(pattern=None, multi=None, abd=False, sports=False, anime_type=None):
        result = naming.test_name(pattern, try_int(multi, None), abd, sports, try_int(anime_type, None))
        result = os.path.join(result["dir"], result["name"])

        return result

    @staticmethod
    def isNamingValid(pattern=None, multi=None, abd=False, sports=False, anime_type=None):
        if not pattern:
            return "invalid"

        # air by date shows just need one check, we don't need to worry about season folders
        if abd:
            is_valid = naming.check_valid_abd_naming(pattern)
            require_season_folders = False

        # sport shows just need one check, we don't need to worry about season folders
        elif sports:
            is_valid = naming.check_valid_sports_naming(pattern)
            require_season_folders = False

        else:
            # check validity of single and multi ep cases for the whole path
            is_valid = naming.check_valid_naming(pattern, try_int(multi, None), try_int(anime_type, None))

            # check validity of single and multi ep cases for only the file name
            require_season_folders = naming.check_force_season_folders(pattern, try_int(multi, None), try_int(anime_type, None))

        if is_valid and not require_season_folders:
            return "valid"
        elif is_valid and require_season_folders:
            return "seasonfolders"
        else:
            return "invalid"

    @staticmethod
    def isRarSupported():
        """
        Test Unpacking Support: - checks if unrar is installed and accesible
        """
        check = config.change_unrar_tool(settings.UNRAR_TOOL, settings.UNAR_TOOL, settings.BSDTAR_TOOL)
        if not check:
            logger.warning("Looks like unrar is not installed, check failed")
        return ("not supported", "supported")[check]



import abc

from sickchill import settings


class Indexer(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def __init__(self):
        self.name = "Generic"
        self.slug = "generic"

        self.language = settings.INDEXER_DEFAULT_LANGUAGE
        self.indexer = settings.INDEXER_DEFAULT
        self.timeout = settings.INDEXER_TIMEOUT

    @abc.abstractmethod
    def search(self, name, language=None, exact=False, indexer_id=False):
        raise NotImplementedError

    @abc.abstractmethod
    def get_series_by_id(self, indexerid, language=None):
        raise NotImplementedError

    @abc.abstractmethod
    def get_series_by_name(self, indexerid, language=None):
        raise NotImplementedError

    @abc.abstractmethod
    def episodes(self, show, season):
        raise NotImplementedError

    @abc.abstractmethod
    def episode(self, show, season, episode):
        raise NotImplementedError

    @property
    def languages(self):
        raise NotImplementedError

    @property
    def lang_dict(self):
        raise NotImplementedError



import os
import threading
import time
import traceback

from sickchill import logger, settings

from . import common, config, generic_queue
from .processTV import log_helper, process_dir

MANUAL_POST_PROCESS = 120
AUTO_POST_PROCESS = 100

# TODO: Add html to the server status page showing processing tasks in the queue
# TODO: Add html to the management section to pause/unpause the processing queue


class PostProcessor(object):
    def __init__(self):
        self.lock = threading.Lock()
        self.amActive = False

    def run(self, force=False):
        """
        Runs the postprocessor

        :param force: Forces postprocessing run
        :return: Returns when done without a return state/code
        """
        self.amActive = True
        settings.postProcessorTaskScheduler.action.add_item(settings.TV_DOWNLOAD_DIR, force=force)
        self.amActive = False

    def __del__(self):
        pass


class ProcessingQueue(generic_queue.GenericQueue):
    """
    Queue to handle multiple post processing tasks
    """

    def __init__(self):
        """

        :rtype: object
        """
        super().__init__()
        self.queue_name = "POSTPROCESSOR"

    def find_in_queue(self, directory, filename, mode):
        """
        Finds any item in the queue with the given directory and mode pair
        :param directory: directory to be processed by the task
        :param filename: filename/nzbname/release_name to be processed by the task
        :param mode: processing type, auto/manual
        :return: instance of PostProcessorTask or None
        """
        for cur_item in self.queue + [self.currentItem]:
            if isinstance(cur_item, PostProcessorTask) and cur_item.matches(directory, filename, mode):
                return cur_item
        return None

    @property
    def is_paused(self):
        """
        Shows if the post processing queue is paused
        :return: bool
        """
        return self.min_priority == generic_queue.QueuePriorities.HIGH

    @property
    def pause(self):
        """
        Pause the post processing queue
        """
        self.min_priority = generic_queue.QueuePriorities.HIGH
        return True

    @property
    def unpause(self):
        """
        Unpause the processing queue
        """
        self.min_priority = 0

    def queue_length(self):
        """
        Returns a dict showing how many auto and manual tasks are in the queue
        :return: dict
        """
        length = {"auto": 0, "manual": 0}
        for cur_item in self.queue + [self.currentItem]:
            if isinstance(cur_item, PostProcessorTask):
                if cur_item.mode == "auto":
                    length["auto"] += 1
                else:
                    length["manual"] += 1
        return length

    def add_item(self, directory, filename=None, method=None, force=False, is_priority=None, delete=None, failed=False, mode="auto", force_next=False):
        """
        Adds a processing task to the queue
        :param directory: directory to process
        :param filename: release/nzb name if available
        :param method: processing method, copy/move/symlink/link
        :param force: force overwriting of existing files regardless of quality
        :param is_priority: whether to replace the file even if it exists at higher quality
        :param delete: delete files and folders after they are processed (always happens with move and auto combination)
        :param failed: mark downloads as failed if they fail to process
        :param mode: processing type: auto/manual
        :param force_next: wait until the current item in the queue is finished, acquire the lock and process this task now, so we can return the result
        :return: string indicating success or failure
        """
        replacements = dict(mode=mode.title(), info=filename or directory)
        if not directory:
            return log_helper("{mode} post-processing attempted but directory is not set: {info}".format(**replacements), logger.WARNING)

        # if not os.path.isdir(directory):
        #     return log_helper("{mode} post-processing attempted but directory doesn't exist: {info}".format(
        #         **replacements), logger.WARNING)

        if not os.path.isabs(directory):
            return log_helper(
                "{mode} post-processing attempted but directory is relative (and probably not what you really want to process): {info}".format(**replacements),
                logger.WARNING,
            )

        item = self.find_in_queue(directory, filename, mode)

        if not delete:
            delete = (False, (not settings.NO_DELETE, True)[method == "move"])[mode == "auto"]

        if item:
            if self.currentItem == item:
                return log_helper("{info} is already being processed right now, please wait until it completes before trying again".format(**replacements))

            item.set_params(directory, filename, method, force, is_priority, delete, failed, mode)
            message = log_helper("A task for {info} was already in the processing queue, updating the settings for that task".format(**replacements))
            return message + r"<br\><span class='hidden'>Processing succeeded</span>"
        else:
            item = PostProcessorTask(directory, filename, method, force, is_priority, delete, failed, mode)
            if force_next:
                with self.lock:
                    item.run()  # Non threaded, but with queue lock
                    message = item.last_result
                return message
            else:
                super().add_item(item)
                message = log_helper("{mode} post processing task for {info} was added to the queue".format(**replacements))
                return message + r"<br\><span class='hidden'>Processing succeeded</span>"


class PostProcessorTask(generic_queue.QueueItem):
    """
    Processing task
    """

    def __init__(self, directory, filename=None, method=None, force=False, is_priority=None, delete=False, failed=False, mode="auto"):
        """
        :param directory: directory to process
        :param filename: release/nzb name if available
        :param method: processing method, copy/move/symlink/link
        :param force: force overwriting of existing files regardless of quality
        :param is_priority: whether to replace the file even if it exists at higher quality
        :param delete: delete files and folders after they are processed (always happens with move and auto combination)
        :param failed: mark downloads as failed if they fail to process
        :param mode: processing type: auto/manual
        :return: None
        """
        super().__init__("{mode}".format(mode=mode.title()), (MANUAL_POST_PROCESS, AUTO_POST_PROCESS)[mode == "auto"])

        self.directory = directory
        self.filename = filename
        self.method = method
        self.force = config.checkbox_to_value(force)
        self.is_priority = config.checkbox_to_value(is_priority)
        self.delete = config.checkbox_to_value(delete)
        self.failed = config.checkbox_to_value(failed)
        self.mode = mode

        self.priority = (generic_queue.QueuePriorities.HIGH, generic_queue.QueuePriorities.NORMAL)[mode == "auto"]

        self.last_result = None

    def set_params(self, directory, filename=None, method=None, force=False, is_priority=None, delete=False, failed=False, mode="auto"):
        """
        Adjust settings for a task that is already in the queue
        :param directory: directory to process
        :param filename: release/nzb name if available
        :param method: processing method, copy/move/symlink/link
        :param force: force overwriting of existing files regardless of quality
        :param is_priority: whether to replace the file even if it exists at higher quality
        :param delete: delete files and folders after they are processed (always happens with move and auto combination)
        :param failed: mark downloads as failed if they fail to process
        :param mode: processing type: auto/manual
        :return: None
        """
        self.directory = directory
        self.filename = filename
        self.method = method
        self.force = config.checkbox_to_value(force)
        self.is_priority = config.checkbox_to_value(is_priority)
        self.delete = config.checkbox_to_value(delete)
        self.failed = config.checkbox_to_value(failed)
        self.mode = mode

    @property
    def info(self):
        return self.directory, self.filename, self.mode

    def matches(self, directory, filename, mode):
        return self.info == (directory, filename, mode)

    def run(self):
        """
        Runs the task
        :return: None
        """
        super().run()

        # noinspection PyBroadException
        try:
            logger.info("Beginning {mode} post processing task: {info}".format(mode=self.mode, info=self.filename or self.directory))
            self.last_result = process_dir(
                process_path=self.directory,
                release_name=self.filename,
                process_method=self.method,
                force=self.force,
                is_priority=self.is_priority,
                delete_on=self.delete,
                failed=self.failed,
                mode=self.mode,
            )
            logger.info("{mode} post processing task for {info} completed".format(mode=self.mode.title(), info=self.filename or self.directory))

            # give the CPU a break
            time.sleep(common.cpu_presets[settings.CPU_PRESET])
        except Exception:
            logger.debug(traceback.format_exc())

        super().finish()
        self.finish()



from urllib.parse import urljoin

import requests

from sickchill import logger, settings


class Notifier(object):
    def __init__(self):
        self.session = None

    def __make_session(self, emby_apikey=None):
        if not self.session:
            self.session = requests.Session()

        self.session.headers.update({"X-Emby-Token": emby_apikey or settings.EMBY_APIKEY, "Content-Type": "application/json"})
        return self.session

    def _notify_emby(self, message, host=None, emby_apikey=None):
        """Handles notifying Emby host via HTTP API

        Returns:
            Returns True for no issue or False if there was an error

        """
        url = urljoin(host or settings.EMBY_HOST, "emby/Notifications/Admin")
        params = {"Name": "SickChill", "Description": message, "ImageUrl": settings.LOGO_URL}

        try:
            session = self.__make_session(emby_apikey)
            response = session.get(url, params=params)
            if response:
                logger.debug("EMBY: HTTP response: {0}".format(response.text.replace("\n", "")))
            response.raise_for_status()

            return True
        except requests.exceptions.RequestException as error:
            logger.warning(f"EMBY: Warning: Could not contact Emby at {url} {error}")
            return False

    ##############################################################################
    # Public functions
    ##############################################################################

    def test_notify(self, host, emby_apikey):
        return self._notify_emby("This is a test notification from SickChill", host, emby_apikey)

    def update_library(self, show=None):
        """Handles updating the Emby Media Server host via HTTP API

        Returns:
            Returns True for no issue or False if there was an error

        """

        if settings.USE_EMBY:
            if not settings.EMBY_HOST:
                logger.debug("EMBY: No host specified, check your settings")
                return False

            params = {}
            if show:
                params.update({"TvdbId": show.indexerid})
                # Endpoint emby/Library/Series/Added is deprecated http://swagger.emby.media/?staticview=true#/LibraryService/postLibrarySeriesAdded
                url = urljoin(settings.EMBY_HOST, "emby/Library/Series/Added")
            else:
                url = urljoin(settings.EMBY_HOST, "emby/Library/Refresh")

            try:
                session = self.__make_session()
                response = session.post(url, params=params)
                response.raise_for_status()
                logger.debug("EMBY: HTTP status: {0}, response: {1}".format(response.status_code, response.text.replace("\n", "")))
                return True

            except requests.exceptions.RequestException as error:
                logger.warning(f"EMBY: Warning: Could not contact Emby at {url} {error}")

                return False



import re
from urllib.parse import urljoin

from requests.utils import dict_from_cookiejar

from sickchill import logger
from sickchill.helper.common import convert_size, try_int
from sickchill.oldbeard import tvcache
from sickchill.oldbeard.bs4_parser import BS4Parser
from sickchill.providers.torrent.TorrentProvider import TorrentProvider


class Provider(TorrentProvider):
    def __init__(self):

        # Provider Init
        super().__init__("ArcheTorrent")

        # Credentials
        self.username = None
        self.password = None

        # Torrent Stats
        self.minseed = 0
        self.minleech = 0

        # Freelech
        self.freeleech = False

        # URLs
        self.url = "https://www.archetorrent.com/"
        self.urls = {
            "login": urljoin(self.url, "account-login.php"),
            "search": urljoin(self.url, "torrents-search.php"),
            "download": urljoin(self.url, "download.php"),
        }

        # Proper Strings
        self.proper_strings = ["PROPER"]

        # Cache
        self.cache = tvcache.TVCache(self, min_time=15)

    def login(self):
        if any(dict_from_cookiejar(self.session.cookies).values()):
            return True

        login_params = {"username": self.username, "password": self.password, "returnto": "/index.php"}

        response = self.get_url(self.urls["login"], post_data=login_params, returns="text")
        if not response:
            logger.warning("Unable to connect to provider")
            return False

        search = self.get_url(self.urls["search"])

        if not search or not re.search("torrents.php", search):
            logger.warning("Invalid username or password. Check your settings")
            return False

        return True

    def search(self, search_strings, age=0, ep_obj=None):
        results = []
        if not self.login():
            return results

        freeleech = "2" if self.freeleech else "0"

        # Search Params
        # c59=1&c73=1&c5=1&c41=1&c60=1&c66=1&c65=1&c67=1&c62=1&c64=1&c61=1&search=Good+Behavior+S01E01&cat=0&incldead=0&freeleech=0&lang=0
        search_params = {
            "c5": "1",  # Category: Series - DVDRip
            "c41": "1",  # Category: Series - HD
            "c60": "1",  # Category: Series - Pack TV
            "c62": "1",  # Category: Series - BDRip
            "c64": "1",  # Category: Series - VOSTFR
            "c65": "1",  # Category: Series - TV 720p
            "c66": "1",  # Category: Series - TV 1080p
            "c67": "1",  # Category: Series - Pack TV HD
            "c73": "1",  # Category: Anime
            "incldead": "0",  # Include dead torrent - 0: off 1: yes 2: only dead
            "freeleech": freeleech,  # Only freeleech torrent - 0: off 1: no freeleech 2: Only freeleech
            "lang": "0",  # Langugage - 0: off 1: English 2: French ....
        }

        # Units
        units = ["B", "KB", "MB", "GB", "TB", "PB"]

        for mode in search_strings:
            items = []
            logger.debug(_("Search Mode: {mode}".format(mode=mode)))

            for search_string in {*search_strings[mode]}:
                logger.debug("Search String: {0} for mode {1}".format(search_strings[mode], mode))
                if mode != "RSS":
                    logger.debug("Search string: {0}".format(search_string))

                search_params["search"] = re.sub(r"[()]", "", search_string)
                data = self.get_url(self.urls["search"], params=search_params, returns="text")
                if not data:
                    continue

                with BS4Parser(data, "html5lib") as html:
                    torrent_table = html.find(class_="ttable_headinner")
                    torrent_rows = torrent_table("tr") if torrent_table else []

                    # Continue only if at least one Release is found
                    if len(torrent_rows) < 2:
                        logger.debug("Data returned from provider does not contain any torrents")
                        continue

                    # Catégorie, Release, Date, DL, Size, C, S, L
                    labels = [label.get_text(strip=True) for label in torrent_rows[0]("th")]

                    # Skip column headers
                    for result in torrent_rows[1:]:
                        cells = result("td")
                        if len(cells) < len(labels):
                            continue

                        try:
                            id = re.search("id=([0-9]+)", cells[labels.index("Nom")].find("a")["href"]).group(1)
                            title = cells[labels.index("Nom")].get_text(strip=True)
                            download_url = urljoin(self.urls["download"], "?id={0}&name={1}".format(id, title))
                            if not all([title, download_url]):
                                continue

                            seeders = try_int(cells[labels.index("S")].get_text(strip=True))
                            leechers = try_int(cells[labels.index("L")].get_text(strip=True))

                            # Filter unseeded torrent
                            if seeders < self.minseed or leechers < self.minleech:
                                if mode != "RSS":
                                    logger.debug(
                                        "Discarding torrent because it doesn't meet the minimum seeders or leechers: {0} (S:{1} L:{2})".format(
                                            title, seeders, leechers
                                        )
                                    )
                                continue

                            size_index = labels.index("Size") if "Size" in labels else labels.index("Taille")
                            torrent_size = cells[size_index].get_text()
                            size = convert_size(torrent_size, units=units) or -1

                            item = {"title": title, "link": download_url, "size": size, "seeders": seeders, "leechers": leechers, "hash": ""}
                            if mode != "RSS":
                                logger.debug(
                                    _(
                                        "Found result: {title} with {seeders} seeders and {leechers} leechers".format(
                                            title=title, seeders=seeders, leechers=leechers
                                        )
                                    )
                                )

                            items.append(item)
                        except Exception:
                            continue

            # For each search mode sort all the items by seeders if available
            items.sort(key=lambda d: try_int(d.get("seeders", 0)), reverse=True)
            results += items

        return results



from collections.abc import MutableMapping


class NumDict(MutableMapping):
    """
    NumDict() -> new empty dictionary

    NumDict(mapping) -> new dictionary initialized from a mapping object's
        (key, value) pairs

    NumDict(iterable) -> new dictionary initialized as if via:
        d = {}
        for k, v in iterable:
            d[k] = v

    NumDict(**kwargs) -> TypeError - key words cannot be numeric

    All keys must be numeric or None
    """

    def __init__(self, iterable=None, **kwargs):
        self.data = {}
        iterable = kwargs.pop("dict", None) if iterable is None else iterable
        if iterable is not None:
            self.update(iterable)
        if kwargs:
            self.update(kwargs)

    def __len__(self):
        return len(self.data)

    # noinspection PyCallingNonCallable
    def __getitem__(self, key):
        key = self.numeric(key)
        if key in self.data:
            return self.data[key]
        if hasattr(self.__class__, "__missing__"):
            # noinspection PyUnresolvedReferences
            return self.__class__.__missing__(self, key)
        raise KeyError(key)

    def __setitem__(self, key, item):
        try:
            key = self.numeric(key)
        except KeyError:
            raise TypeError(key)
        self.data[key] = item

    def __delitem__(self, key):
        key = self.numeric(key)
        del self.data[key]

    def __iter__(self):
        return iter(self.data)

    def __contains__(self, key):
        try:
            key = self.numeric(key)
            return key in self.data
        except KeyError:
            return False

    def __repr__(self):
        return repr(self.data)

    def has_key(self, key):
        """
        DEPRECATED: Check for existence of key

        :param key: A numeric key
        :return: True if key is found, else False
        """
        return key in self

    def copy(self):
        """
        Create a copy of a NumDict
        :return: A copy
        """
        if self.__class__ is NumDict:
            return NumDict(self.data.copy())
        import copy

        data = self.data
        try:
            self.data = {}
            c = copy.copy(self)
        finally:
            self.data = data
        c.update(self)
        return c

    @classmethod
    def fromkeys(cls, iterable, value=None):
        """
        Build a NumDict from a dictionary

        :param iterable:
        :param value:
        :return:
        """
        d = cls()
        for key in iterable:
            key = cls.numeric(key)
            d[key] = value
        return d

    @staticmethod
    def numeric(key):
        """
        Converts a key to its numeric representation

        :param key: numeric dict key
        :raise KeyError: if key can't be converted to an integer
        :return: a numeric key
        :rtype: int
        """
        if key is None:
            key = 0
        try:
            return int(key)
        except (TypeError, ValueError):
            if key is not None:
                raise KeyError(key)
